import json

from vyper import compiler
from vyper.ast.nodes import NODE_SRC_ATTRIBUTES
from vyper.ast.parse import parse_to_ast
from vyper.ast.utils import ast_to_dict, dict_to_ast


def get_node_ids(ast_struct, ids=None):
    if ids is None:
        ids = []

    for k, v in ast_struct.items():
        if isinstance(v, dict):
            ids = get_node_ids(v, ids)
        elif isinstance(v, list):
            for x in v:
                ids = get_node_ids(x, ids)
        elif k == "node_id":
            ids.append(v)
        elif v is None or isinstance(v, (str, int)):
            continue
        else:
            raise Exception(f"Unknown ast_struct provided. {k}, {v}")
    return ids


def test_ast_to_dict_node_id():
    code = """
@external
def test() -> int128:
    a: uint256 = 100
    return 123
    """
    dict_out = compiler.compile_code(code, output_formats=["ast_dict"])
    node_ids = get_node_ids(dict_out)

    assert len(node_ids) == len(set(node_ids))


def test_basic_ast():
    code = """
a: int128
    """
    dict_out = compiler.compile_code(code, output_formats=["annotated_ast_dict"], source_id=0)
    assert dict_out["annotated_ast_dict"]["ast"]["body"][0] == {
        "annotation": {
            "ast_type": "Name",
            "col_offset": 3,
            "end_col_offset": 9,
            "end_lineno": 2,
            "id": "int128",
            "lineno": 2,
            "node_id": 4,
            "src": "4:6:0",
        },
        "ast_type": "VariableDecl",
        "col_offset": 0,
        "end_col_offset": 9,
        "end_lineno": 2,
        "is_constant": False,
        "is_immutable": False,
        "is_public": False,
        "is_transient": False,
        "lineno": 2,
        "node_id": 1,
        "src": "1:9:0",
        "target": {
            "ast_type": "Name",
            "col_offset": 0,
            "end_col_offset": 1,
            "end_lineno": 2,
            "id": "a",
            "lineno": 2,
            "node_id": 2,
            "src": "1:1:0",
            "type": {"bits": 128, "is_signed": True, "name": "int128", "typeclass": "integer"},
        },
        "type": {"bits": 128, "is_signed": True, "name": "int128", "typeclass": "integer"},
        "value": None,
    }


def test_implements_ast():
    code = """
interface Foo:
    def foo() -> uint256: view

implements: Foo

@external
@view
def foo() -> uint256:
    return 1
    """
    dict_out = compiler.compile_code(code, output_formats=["ast_dict"], source_id=0)
    assert dict_out["ast_dict"]["ast"]["body"][1] == {
        "col_offset": 0,
        "annotation": {
            "col_offset": 12,
            "end_col_offset": 15,
            "node_id": 12,
            "src": "60:3:0",
            "ast_type": "Name",
            "end_lineno": 5,
            "lineno": 5,
            "id": "Foo",
        },
        "end_col_offset": 15,
        "node_id": 9,
        "src": "48:15:0",
        "ast_type": "ImplementsDecl",
        "end_lineno": 5,
        "lineno": 5,
    }


def test_dict_to_ast():
    code = """
@external
def test() -> int128:
    a: uint256 = 100
    b: int128 = -22
    c: decimal = -3.3133700
    d: Bytes[11] = b"oh hai mark"
    e: Bytes[1] = 0b01010101
    f: Bytes[88] = b"\x01\x70"
    g: String[100] = "  baka baka   "
    h: address = 0xdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef
    i: bool = False
    return 123
    """

    original_ast = parse_to_ast(code)
    out_dict = ast_to_dict(original_ast)
    out_json = json.dumps(out_dict)
    new_dict = json.loads(out_json)
    new_ast = dict_to_ast(new_dict)

    assert new_ast == original_ast


# strip source annotations like lineno, we don't care for inspecting
# the analysis result
def _strip_source_annotations(dict_node, to_strip):
    if isinstance(dict_node, dict):
        for k in list(dict_node.keys()):
            if k in to_strip:
                del dict_node[k]
                continue
            if "decl_node" not in k:
                _strip_source_annotations(dict_node[k], to_strip)
    elif isinstance(dict_node, list):
        for child in dict_node:
            _strip_source_annotations(child, to_strip)


def test_output_type_info(make_input_bundle, chdir_tmp_path):
    # test type info is output in the ast dict
    # test different, complex types, and test import info is also output
    lib1 = """
struct Foo:
    x: uint256

event Bar:
    pass

struct Baz:
    x: decimal
    y: Bytes[20]
    z: String[32]
    w: uint256
    u: address

interface Qux:
    def return_tuple() -> (Foo[1], uint256): nonpayable

foo_var: Foo
sarray_var: Foo[1]
darray_var: DynArray[Foo, 5]
interface_var: Qux

hashmap_var: HashMap[address, Foo]

sarray_var2: uint256[2]
darray_var2: DynArray[uint256, 5]

@internal
def foo():
    t: uint256 = max_value(uint256)
    u: int24 = empty(int24)

    self.foo_var = empty(Foo)
    self.sarray_var[0] = empty(Foo)
    self.darray_var[1] = empty(Foo)

    self.sarray_var, t = extcall self.interface_var.return_tuple()

@external
def bar():
    s: bytes24 = empty(bytes24)
    """

    main = """
import lib1

initializes: lib1

@internal
def foo():
    lib1.foo()
    log lib1.Bar()
    s: lib1.Foo = empty(lib1.Foo)
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1, "main.vy": main})

    lib1_file = input_bundle.load_file("lib1.vy")
    out = compiler.compile_from_file_input(
        lib1_file, input_bundle=input_bundle, output_formats=["annotated_ast_dict"]
    )
    lib1_ast = out["annotated_ast_dict"]["ast"]
    lib1_sha256sum = lib1_ast.pop("source_sha256sum")
    assert lib1_sha256sum == lib1_file.sha256sum
    to_strip = NODE_SRC_ATTRIBUTES + ("resolved_path", "variable_reads", "variable_writes")
    _strip_source_annotations(lib1_ast, to_strip=to_strip)

    main_file = input_bundle.load_file("main.vy")
    out = compiler.compile_from_file_input(
        main_file, input_bundle=input_bundle, output_formats=["annotated_ast_dict"]
    )
    main_ast = out["annotated_ast_dict"]["ast"]
    main_sha256sum = main_ast.pop("source_sha256sum")
    assert main_sha256sum == main_file.sha256sum
    _strip_source_annotations(main_ast, to_strip=to_strip)

    # TODO: would be nice to refactor this into bunch of small test cases
    assert main_ast == {
        "ast_type": "Module",
        "body": [
            {
                "alias": None,
                "ast_type": "Import",
                "import_info": {
                    "alias": "lib1",
                    "file_sha256sum": lib1_file.sha256sum,
                    "path": "lib1.vy",
                    "qualified_module_name": "lib1",
                    "source_id": 0,
                },
                "name": "lib1",
                "node_id": 1,
            },
            {
                "annotation": {"ast_type": "Name", "id": "lib1", "node_id": 6},
                "ast_type": "InitializesDecl",
                "node_id": 3,
            },
            {
                "args": {
                    "args": [],
                    "ast_type": "arguments",
                    "default": None,
                    "defaults": [],
                    "node_id": 9,
                },
                "ast_type": "FunctionDef",
                "body": [
                    {
                        "ast_type": "Expr",
                        "node_id": 10,
                        "value": {
                            "args": [],
                            "ast_type": "Call",
                            "func": {
                                "ast_type": "Attribute",
                                "attr": "foo",
                                "node_id": 12,
                                "type": {
                                    "name": "foo",
                                    "type_decl_node": {"node_id": 119, "source_id": 0},
                                    "typeclass": "contract_function",
                                },
                                "value": {
                                    "ast_type": "Name",
                                    "id": "lib1",
                                    "node_id": 13,
                                    "type": {
                                        "name": "lib1.vy",
                                        "type_decl_node": {"node_id": 0, "source_id": 0},
                                        "typeclass": "module",
                                    },
                                },
                            },
                            "keywords": [],
                            "node_id": 11,
                            "type": {"name": "(void)"},
                        },
                    },
                    {
                        "ast_type": "Log",
                        "node_id": 17,
                        "type": {
                            "name": "Bar",
                            "type_decl_node": {"node_id": 7, "source_id": 0},
                            "typeclass": "event",
                        },
                        "value": {
                            "args": [],
                            "ast_type": "Call",
                            "func": {
                                "ast_type": "Attribute",
                                "attr": "Bar",
                                "node_id": 19,
                                "type": {
                                    "type_t": {
                                        "name": "Bar",
                                        "type_decl_node": {"node_id": 7, "source_id": 0},
                                        "typeclass": "event",
                                    }
                                },
                                "value": {
                                    "ast_type": "Name",
                                    "id": "lib1",
                                    "node_id": 20,
                                    "type": {
                                        "name": "lib1.vy",
                                        "type_decl_node": {"node_id": 0, "source_id": 0},
                                        "typeclass": "module",
                                    },
                                },
                            },
                            "keywords": [],
                            "node_id": 18,
                            "type": {"name": "(void)"},
                        },
                    },
                    {
                        "annotation": {
                            "ast_type": "Attribute",
                            "attr": "Foo",
                            "node_id": 26,
                            "value": {"ast_type": "Name", "id": "lib1", "node_id": 27},
                        },
                        "ast_type": "AnnAssign",
                        "node_id": 23,
                        "target": {
                            "ast_type": "Name",
                            "id": "s",
                            "node_id": 24,
                            "type": {"name": "Foo", "typeclass": "struct"},
                        },
                        "value": {
                            "args": [
                                {
                                    "ast_type": "Attribute",
                                    "attr": "Foo",
                                    "node_id": 33,
                                    "type": {"type_t": {"name": "Foo", "typeclass": "struct"}},
                                    "value": {
                                        "ast_type": "Name",
                                        "id": "lib1",
                                        "node_id": 34,
                                        "type": {
                                            "name": "lib1.vy",
                                            "type_decl_node": {"node_id": 0, "source_id": 0},
                                            "typeclass": "module",
                                        },
                                    },
                                }
                            ],
                            "ast_type": "Call",
                            "func": {
                                "ast_type": "Name",
                                "id": "empty",
                                "node_id": 31,
                                "type": {"name": "empty", "typeclass": "builtin_function"},
                            },
                            "keywords": [],
                            "node_id": 30,
                            "type": {"name": "Foo", "typeclass": "struct"},
                        },
                    },
                ],
                "decorator_list": [{"ast_type": "Name", "id": "internal", "node_id": 37}],
                "doc_string": None,
                "name": "foo",
                "node_id": 8,
                "pos": None,
                "returns": None,
            },
        ],
        "doc_string": None,
        "name": None,
        "node_id": 0,
        "path": "main.vy",
        "source_id": 1,
        "type": {
            "name": "main.vy",
            "type_decl_node": {"node_id": 0, "source_id": 1},
            "typeclass": "module",
        },
    }

    # TODO: would be nice to refactor this into bunch of small test cases
    # TODO: write the test in a way which makes the links between nodes
    # clearer
    assert lib1_ast == {
        "ast_type": "Module",
        "body": [
            {
                "ast_type": "StructDef",
                "body": [
                    {
                        "annotation": {"ast_type": "Name", "id": "uint256", "node_id": 5},
                        "ast_type": "AnnAssign",
                        "node_id": 2,
                        "target": {"ast_type": "Name", "id": "x", "node_id": 3},
                        "value": None,
                    }
                ],
                "doc_string": None,
                "name": "Foo",
                "node_id": 1,
            },
            {
                "ast_type": "EventDef",
                "body": [{"ast_type": "Pass", "node_id": 8}],
                "doc_string": None,
                "name": "Bar",
                "node_id": 7,
            },
            {
                "ast_type": "StructDef",
                "body": [
                    {
                        "annotation": {"ast_type": "Name", "id": "decimal", "node_id": 13},
                        "ast_type": "AnnAssign",
                        "node_id": 10,
                        "target": {"ast_type": "Name", "id": "x", "node_id": 11},
                        "value": None,
                    },
                    {
                        "annotation": {
                            "ast_type": "Subscript",
                            "node_id": 18,
                            "slice": {"ast_type": "Int", "node_id": 21, "value": 20},
                            "value": {"ast_type": "Name", "id": "Bytes", "node_id": 19},
                        },
                        "ast_type": "AnnAssign",
                        "node_id": 15,
                        "target": {"ast_type": "Name", "id": "y", "node_id": 16},
                        "value": None,
                    },
                    {
                        "annotation": {
                            "ast_type": "Subscript",
                            "node_id": 26,
                            "slice": {"ast_type": "Int", "node_id": 29, "value": 32},
                            "value": {"ast_type": "Name", "id": "String", "node_id": 27},
                        },
                        "ast_type": "AnnAssign",
                        "node_id": 23,
                        "target": {"ast_type": "Name", "id": "z", "node_id": 24},
                        "value": None,
                    },
                    {
                        "annotation": {"ast_type": "Name", "id": "uint256", "node_id": 34},
                        "ast_type": "AnnAssign",
                        "node_id": 31,
                        "target": {"ast_type": "Name", "id": "w", "node_id": 32},
                        "value": None,
                    },
                    {
                        "annotation": {"ast_type": "Name", "id": "address", "node_id": 39},
                        "ast_type": "AnnAssign",
                        "node_id": 36,
                        "target": {"ast_type": "Name", "id": "u", "node_id": 37},
                        "value": None,
                    },
                ],
                "doc_string": None,
                "name": "Baz",
                "node_id": 9,
            },
            {
                "ast_type": "InterfaceDef",
                "body": [
                    {
                        "args": {
                            "args": [],
                            "ast_type": "arguments",
                            "default": None,
                            "defaults": [],
                            "node_id": 43,
                        },
                        "ast_type": "FunctionDef",
                        "body": [
                            {
                                "ast_type": "Expr",
                                "node_id": 44,
                                "value": {"ast_type": "Name", "id": "nonpayable", "node_id": 45},
                            }
                        ],
                        "decorator_list": [],
                        "doc_string": None,
                        "name": "return_tuple",
                        "node_id": 42,
                        "pos": None,
                        "returns": {
                            "ast_type": "Tuple",
                            "elements": [
                                {
                                    "ast_type": "Subscript",
                                    "node_id": 48,
                                    "slice": {"ast_type": "Int", "node_id": 51, "value": 1},
                                    "value": {"ast_type": "Name", "id": "Foo", "node_id": 49},
                                },
                                {"ast_type": "Name", "id": "uint256", "node_id": 53},
                            ],
                            "node_id": 47,
                        },
                    }
                ],
                "doc_string": None,
                "name": "Qux",
                "node_id": 41,
            },
            {
                "annotation": {"ast_type": "Name", "id": "Foo", "node_id": 59},
                "ast_type": "VariableDecl",
                "is_constant": False,
                "is_immutable": False,
                "is_public": False,
                "is_transient": False,
                "node_id": 56,
                "target": {
                    "ast_type": "Name",
                    "id": "foo_var",
                    "node_id": 57,
                    "type": {"name": "Foo", "typeclass": "struct"},
                },
                "type": {"name": "Foo", "typeclass": "struct"},
                "value": None,
            },
            {
                "annotation": {
                    "ast_type": "Subscript",
                    "node_id": 64,
                    "slice": {"ast_type": "Int", "node_id": 67, "value": 1},
                    "value": {"ast_type": "Name", "id": "Foo", "node_id": 65},
                },
                "ast_type": "VariableDecl",
                "is_constant": False,
                "is_immutable": False,
                "is_public": False,
                "is_transient": False,
                "node_id": 61,
                "target": {
                    "ast_type": "Name",
                    "id": "sarray_var",
                    "node_id": 62,
                    "type": {
                        "length": 1,
                        "name": "$SArray",
                        "typeclass": "static_array",
                        "value_type": {"name": "Foo", "typeclass": "struct"},
                    },
                },
                "type": {
                    "length": 1,
                    "name": "$SArray",
                    "typeclass": "static_array",
                    "value_type": {"name": "Foo", "typeclass": "struct"},
                },
                "value": None,
            },
            {
                "annotation": {
                    "ast_type": "Subscript",
                    "node_id": 72,
                    "slice": {
                        "ast_type": "Tuple",
                        "elements": [
                            {"ast_type": "Name", "id": "Foo", "node_id": 76},
                            {"ast_type": "Int", "node_id": 78, "value": 5},
                        ],
                        "node_id": 75,
                    },
                    "value": {"ast_type": "Name", "id": "DynArray", "node_id": 73},
                },
                "ast_type": "VariableDecl",
                "is_constant": False,
                "is_immutable": False,
                "is_public": False,
                "is_transient": False,
                "node_id": 69,
                "target": {
                    "ast_type": "Name",
                    "id": "darray_var",
                    "node_id": 70,
                    "type": {
                        "length": 5,
                        "name": "DynArray",
                        "typeclass": "dynamic_array",
                        "value_type": {"name": "Foo", "typeclass": "struct"},
                    },
                },
                "type": {
                    "length": 5,
                    "name": "DynArray",
                    "typeclass": "dynamic_array",
                    "value_type": {"name": "Foo", "typeclass": "struct"},
                },
                "value": None,
            },
            {
                "annotation": {"ast_type": "Name", "id": "Qux", "node_id": 84},
                "ast_type": "VariableDecl",
                "is_constant": False,
                "is_immutable": False,
                "is_public": False,
                "is_transient": False,
                "node_id": 81,
                "target": {
                    "ast_type": "Name",
                    "id": "interface_var",
                    "node_id": 82,
                    "type": {
                        "name": "Qux",
                        "type_decl_node": {"node_id": 41, "source_id": 0},
                        "typeclass": "interface",
                    },
                },
                "type": {
                    "name": "Qux",
                    "type_decl_node": {"node_id": 41, "source_id": 0},
                    "typeclass": "interface",
                },
                "value": None,
            },
            {
                "annotation": {
                    "ast_type": "Subscript",
                    "node_id": 89,
                    "slice": {
                        "ast_type": "Tuple",
                        "elements": [
                            {"ast_type": "Name", "id": "address", "node_id": 93},
                            {"ast_type": "Name", "id": "Foo", "node_id": 95},
                        ],
                        "node_id": 92,
                    },
                    "value": {"ast_type": "Name", "id": "HashMap", "node_id": 90},
                },
                "ast_type": "VariableDecl",
                "is_constant": False,
                "is_immutable": False,
                "is_public": False,
                "is_transient": False,
                "node_id": 86,
                "target": {
                    "ast_type": "Name",
                    "id": "hashmap_var",
                    "node_id": 87,
                    "type": {
                        "key_type": {"name": "address"},
                        "name": "HashMap",
                        "typeclass": "hashmap",
                        "value_type": {"name": "Foo", "typeclass": "struct"},
                    },
                },
                "type": {
                    "key_type": {"name": "address"},
                    "name": "HashMap",
                    "typeclass": "hashmap",
                    "value_type": {"name": "Foo", "typeclass": "struct"},
                },
                "value": None,
            },
            {
                "annotation": {
                    "ast_type": "Subscript",
                    "node_id": 102,
                    "slice": {"ast_type": "Int", "node_id": 105, "value": 2},
                    "value": {"ast_type": "Name", "id": "uint256", "node_id": 103},
                },
                "ast_type": "VariableDecl",
                "is_constant": False,
                "is_immutable": False,
                "is_public": False,
                "is_transient": False,
                "node_id": 99,
                "target": {
                    "ast_type": "Name",
                    "id": "sarray_var2",
                    "node_id": 100,
                    "type": {
                        "length": 2,
                        "name": "$SArray",
                        "typeclass": "static_array",
                        "value_type": {
                            "bits": 256,
                            "is_signed": False,
                            "name": "uint256",
                            "typeclass": "integer",
                        },
                    },
                },
                "type": {
                    "length": 2,
                    "name": "$SArray",
                    "typeclass": "static_array",
                    "value_type": {
                        "bits": 256,
                        "is_signed": False,
                        "name": "uint256",
                        "typeclass": "integer",
                    },
                },
                "value": None,
            },
            {
                "annotation": {
                    "ast_type": "Subscript",
                    "node_id": 110,
                    "slice": {
                        "ast_type": "Tuple",
                        "elements": [
                            {"ast_type": "Name", "id": "uint256", "node_id": 114},
                            {"ast_type": "Int", "node_id": 116, "value": 5},
                        ],
                        "node_id": 113,
                    },
                    "value": {"ast_type": "Name", "id": "DynArray", "node_id": 111},
                },
                "ast_type": "VariableDecl",
                "is_constant": False,
                "is_immutable": False,
                "is_public": False,
                "is_transient": False,
                "node_id": 107,
                "target": {
                    "ast_type": "Name",
                    "id": "darray_var2",
                    "node_id": 108,
                    "type": {
                        "length": 5,
                        "name": "DynArray",
                        "typeclass": "dynamic_array",
                        "value_type": {
                            "bits": 256,
                            "is_signed": False,
                            "name": "uint256",
                            "typeclass": "integer",
                        },
                    },
                },
                "type": {
                    "length": 5,
                    "name": "DynArray",
                    "typeclass": "dynamic_array",
                    "value_type": {
                        "bits": 256,
                        "is_signed": False,
                        "name": "uint256",
                        "typeclass": "integer",
                    },
                },
                "value": None,
            },
            {
                "args": {
                    "args": [],
                    "ast_type": "arguments",
                    "default": None,
                    "defaults": [],
                    "node_id": 120,
                },
                "ast_type": "FunctionDef",
                "body": [
                    {
                        "annotation": {"ast_type": "Name", "id": "uint256", "node_id": 124},
                        "ast_type": "AnnAssign",
                        "node_id": 121,
                        "target": {
                            "ast_type": "Name",
                            "id": "t",
                            "node_id": 122,
                            "type": {
                                "bits": 256,
                                "is_signed": False,
                                "name": "uint256",
                                "typeclass": "integer",
                            },
                        },
                        "value": {
                            "args": [
                                {
                                    "ast_type": "Name",
                                    "id": "uint256",
                                    "node_id": 129,
                                    "type": {
                                        "type_t": {
                                            "bits": 256,
                                            "is_signed": False,
                                            "name": "uint256",
                                            "typeclass": "integer",
                                        }
                                    },
                                }
                            ],
                            "ast_type": "Call",
                            "func": {
                                "ast_type": "Name",
                                "id": "max_value",
                                "node_id": 127,
                                "type": {"name": "max_value", "typeclass": "builtin_function"},
                            },
                            "keywords": [],
                            "node_id": 126,
                            "type": {
                                "bits": 256,
                                "is_signed": False,
                                "name": "uint256",
                                "typeclass": "integer",
                            },
                        },
                    },
                    {
                        "annotation": {"ast_type": "Name", "id": "int24", "node_id": 134},
                        "ast_type": "AnnAssign",
                        "node_id": 131,
                        "target": {
                            "ast_type": "Name",
                            "id": "u",
                            "node_id": 132,
                            "type": {
                                "bits": 24,
                                "is_signed": True,
                                "name": "int24",
                                "typeclass": "integer",
                            },
                        },
                        "value": {
                            "args": [
                                {
                                    "ast_type": "Name",
                                    "id": "int24",
                                    "node_id": 139,
                                    "type": {
                                        "type_t": {
                                            "bits": 24,
                                            "is_signed": True,
                                            "name": "int24",
                                            "typeclass": "integer",
                                        }
                                    },
                                }
                            ],
                            "ast_type": "Call",
                            "func": {
                                "ast_type": "Name",
                                "id": "empty",
                                "node_id": 137,
                                "type": {"name": "empty", "typeclass": "builtin_function"},
                            },
                            "keywords": [],
                            "node_id": 136,
                            "type": {
                                "bits": 24,
                                "is_signed": True,
                                "name": "int24",
                                "typeclass": "integer",
                            },
                        },
                    },
                    {
                        "ast_type": "Assign",
                        "node_id": 141,
                        "target": {
                            "ast_type": "Attribute",
                            "attr": "foo_var",
                            "node_id": 142,
                            "type": {"name": "Foo", "typeclass": "struct"},
                            "value": {
                                "ast_type": "Name",
                                "id": "self",
                                "node_id": 143,
                                "type": {"name": "self"},
                            },
                        },
                        "value": {
                            "args": [
                                {
                                    "ast_type": "Name",
                                    "id": "Foo",
                                    "node_id": 149,
                                    "type": {"type_t": {"name": "Foo", "typeclass": "struct"}},
                                }
                            ],
                            "ast_type": "Call",
                            "func": {
                                "ast_type": "Name",
                                "id": "empty",
                                "node_id": 147,
                                "type": {"name": "empty", "typeclass": "builtin_function"},
                            },
                            "keywords": [],
                            "node_id": 146,
                            "type": {"name": "Foo", "typeclass": "struct"},
                        },
                    },
                    {
                        "ast_type": "Assign",
                        "node_id": 151,
                        "target": {
                            "ast_type": "Subscript",
                            "node_id": 152,
                            "slice": {
                                "ast_type": "Int",
                                "node_id": 157,
                                "type": {
                                    "bits": 8,
                                    "is_signed": True,
                                    "name": "int8",
                                    "typeclass": "integer",
                                },
                                "value": 0,
                            },
                            "type": {"name": "Foo", "typeclass": "struct"},
                            "value": {
                                "ast_type": "Attribute",
                                "attr": "sarray_var",
                                "node_id": 153,
                                "type": {
                                    "length": 1,
                                    "name": "$SArray",
                                    "typeclass": "static_array",
                                    "value_type": {"name": "Foo", "typeclass": "struct"},
                                },
                                "value": {
                                    "ast_type": "Name",
                                    "id": "self",
                                    "node_id": 154,
                                    "type": {"name": "self"},
                                },
                            },
                        },
                        "value": {
                            "args": [
                                {
                                    "ast_type": "Name",
                                    "id": "Foo",
                                    "node_id": 162,
                                    "type": {"type_t": {"name": "Foo", "typeclass": "struct"}},
                                }
                            ],
                            "ast_type": "Call",
                            "func": {
                                "ast_type": "Name",
                                "id": "empty",
                                "node_id": 160,
                                "type": {"name": "empty", "typeclass": "builtin_function"},
                            },
                            "keywords": [],
                            "node_id": 159,
                            "type": {"name": "Foo", "typeclass": "struct"},
                        },
                    },
                    {
                        "ast_type": "Assign",
                        "node_id": 164,
                        "target": {
                            "ast_type": "Subscript",
                            "node_id": 165,
                            "slice": {
                                "ast_type": "Int",
                                "node_id": 170,
                                "type": {
                                    "bits": 8,
                                    "is_signed": True,
                                    "name": "int8",
                                    "typeclass": "integer",
                                },
                                "value": 1,
                            },
                            "type": {"name": "Foo", "typeclass": "struct"},
                            "value": {
                                "ast_type": "Attribute",
                                "attr": "darray_var",
                                "node_id": 166,
                                "type": {
                                    "length": 5,
                                    "name": "DynArray",
                                    "typeclass": "dynamic_array",
                                    "value_type": {"name": "Foo", "typeclass": "struct"},
                                },
                                "value": {
                                    "ast_type": "Name",
                                    "id": "self",
                                    "node_id": 167,
                                    "type": {"name": "self"},
                                },
                            },
                        },
                        "value": {
                            "args": [
                                {
                                    "ast_type": "Name",
                                    "id": "Foo",
                                    "node_id": 175,
                                    "type": {"type_t": {"name": "Foo", "typeclass": "struct"}},
                                }
                            ],
                            "ast_type": "Call",
                            "func": {
                                "ast_type": "Name",
                                "id": "empty",
                                "node_id": 173,
                                "type": {"name": "empty", "typeclass": "builtin_function"},
                            },
                            "keywords": [],
                            "node_id": 172,
                            "type": {"name": "Foo", "typeclass": "struct"},
                        },
                    },
                    {
                        "ast_type": "Assign",
                        "node_id": 177,
                        "target": {
                            "ast_type": "Tuple",
                            "elements": [
                                {
                                    "ast_type": "Attribute",
                                    "attr": "sarray_var",
                                    "node_id": 179,
                                    "type": {
                                        "length": 1,
                                        "name": "$SArray",
                                        "typeclass": "static_array",
                                        "value_type": {"name": "Foo", "typeclass": "struct"},
                                    },
                                    "value": {
                                        "ast_type": "Name",
                                        "id": "self",
                                        "node_id": 180,
                                        "type": {"name": "self"},
                                    },
                                },
                                {
                                    "ast_type": "Name",
                                    "id": "t",
                                    "node_id": 183,
                                    "type": {
                                        "bits": 256,
                                        "is_signed": False,
                                        "name": "uint256",
                                        "typeclass": "integer",
                                    },
                                },
                            ],
                            "node_id": 178,
                            "type": {"members": {}, "name": "$Tuple", "typeclass": "tuple"},
                        },
                        "value": {
                            "ast_type": "ExtCall",
                            "node_id": 186,
                            "type": {"members": {}, "name": "$Tuple", "typeclass": "tuple"},
                            "value": {
                                "args": [],
                                "ast_type": "Call",
                                "func": {
                                    "ast_type": "Attribute",
                                    "attr": "return_tuple",
                                    "node_id": 188,
                                    "type": {
                                        "name": "return_tuple",
                                        "type_decl_node": {"node_id": 42, "source_id": 0},
                                        "typeclass": "contract_function",
                                    },
                                    "value": {
                                        "ast_type": "Attribute",
                                        "attr": "interface_var",
                                        "node_id": 189,
                                        "type": {
                                            "name": "Qux",
                                            "type_decl_node": {"node_id": 41, "source_id": 0},
                                            "typeclass": "interface",
                                        },
                                        "value": {
                                            "ast_type": "Name",
                                            "id": "self",
                                            "node_id": 190,
                                            "type": {"name": "self"},
                                        },
                                    },
                                },
                                "keywords": [],
                                "node_id": 187,
                                "type": {"members": {}, "name": "$Tuple", "typeclass": "tuple"},
                            },
                        },
                    },
                ],
                "decorator_list": [{"ast_type": "Name", "id": "internal", "node_id": 194}],
                "doc_string": None,
                "name": "foo",
                "node_id": 119,
                "pos": None,
                "returns": None,
            },
            {
                "args": {
                    "args": [],
                    "ast_type": "arguments",
                    "default": None,
                    "defaults": [],
                    "node_id": 197,
                },
                "ast_type": "FunctionDef",
                "body": [
                    {
                        "annotation": {"ast_type": "Name", "id": "bytes24", "node_id": 201},
                        "ast_type": "AnnAssign",
                        "node_id": 198,
                        "target": {
                            "ast_type": "Name",
                            "id": "s",
                            "node_id": 199,
                            "type": {"m": 24, "name": "bytes24", "typeclass": "bytes_m"},
                        },
                        "value": {
                            "args": [
                                {
                                    "ast_type": "Name",
                                    "id": "bytes24",
                                    "node_id": 206,
                                    "type": {
                                        "type_t": {
                                            "m": 24,
                                            "name": "bytes24",
                                            "typeclass": "bytes_m",
                                        }
                                    },
                                }
                            ],
                            "ast_type": "Call",
                            "func": {
                                "ast_type": "Name",
                                "id": "empty",
                                "node_id": 204,
                                "type": {"name": "empty", "typeclass": "builtin_function"},
                            },
                            "keywords": [],
                            "node_id": 203,
                            "type": {"m": 24, "name": "bytes24", "typeclass": "bytes_m"},
                        },
                    }
                ],
                "decorator_list": [{"ast_type": "Name", "id": "external", "node_id": 208}],
                "doc_string": None,
                "name": "bar",
                "node_id": 196,
                "pos": None,
                "returns": None,
            },
        ],
        "doc_string": None,
        "name": None,
        "node_id": 0,
        "path": "lib1.vy",
        "source_id": 0,
        "type": {
            "name": "lib1.vy",
            "type_decl_node": {"node_id": 0, "source_id": 0},
            "typeclass": "module",
        },
    }


def test_output_variable_read_write_analysis(make_input_bundle, chdir_tmp_path):
    # test we output the result of variable read/write correctly
    # note: also tests serialization of structs, strings, static arrays,
    # and type_decl_nodes across modules.
    lib1 = """
struct Foo:
    a: uint256
    b: String[3]
    c: decimal

struct Bar:
    items: Foo[2]

counter: uint256

bars: DynArray[Bar, 10]
    """
    main = """
import lib1

initializes: lib1

counter: uint256

@internal
def foo():
    x: uint256 = lib1.counter
    lib1.counter += 1

@internal
def bar():
    x: uint256 = lib1.counter
    y: uint256 = self.counter
    lib1.counter += 1

@internal
def baz():
    self.bar()  # reads both lib1.counter and self.counter
    self.counter += 1

@internal
def qux():
    lib1.bars = []
    lib1.bars[0] = empty(lib1.Bar)
    lib1.bars[1].items = empty(lib1.Foo[2])

    lib1.bars[1].items[0].a = 1
    lib1.bars[0].items[1].c = 10.0

@internal
def qux2():
    self.qux()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "main.vy": main})

    # preliminaries: main.vy has source_id==0, lib1.vy has source_id==1.
    file = input_bundle.load_file("main.vy")
    assert file.source_id == 0
    assert input_bundle.load_file("lib1.vy").source_id == 1

    out = compiler.compile_from_file_input(
        file, input_bundle=input_bundle, output_formats=["annotated_ast_dict"]
    )
    ast = out["annotated_ast_dict"]["ast"]

    assert ast["path"] == "main.vy"
    assert ast["source_id"] == 0

    _strip_source_annotations(ast, to_strip=NODE_SRC_ATTRIBUTES + ("node_id", "type"))

    foo, bar, baz, qux, qux2 = ast["body"][3:]
    assert foo["name"] == "foo"
    assert foo["body"] == [
        {
            "annotation": {"ast_type": "Name", "id": "uint256"},
            "ast_type": "AnnAssign",
            "target": {"ast_type": "Name", "id": "x"},
            "value": {
                "ast_type": "Attribute",
                "attr": "counter",
                "value": {"ast_type": "Name", "id": "lib1"},
                "variable_reads": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 29, "source_id": 1},
                        "name": "counter",
                    }
                ],
            },
        },
        {
            "ast_type": "AugAssign",
            "op": {"ast_type": "Add"},
            "target": {
                "ast_type": "Attribute",
                "attr": "counter",
                "value": {"ast_type": "Name", "id": "lib1"},
                "variable_reads": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 29, "source_id": 1},
                        "name": "counter",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 29, "source_id": 1},
                        "name": "counter",
                    }
                ],
            },
            "value": {"ast_type": "Int", "value": 1},
        },
    ]

    assert bar["name"] == "bar"
    assert bar["body"] == [
        {
            "annotation": {"ast_type": "Name", "id": "uint256"},
            "ast_type": "AnnAssign",
            "target": {"ast_type": "Name", "id": "x"},
            "value": {
                "ast_type": "Attribute",
                "attr": "counter",
                "value": {"ast_type": "Name", "id": "lib1"},
                "variable_reads": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 29, "source_id": 1},
                        "name": "counter",
                    }
                ],
            },
        },
        {
            "annotation": {"ast_type": "Name", "id": "uint256"},
            "ast_type": "AnnAssign",
            "target": {"ast_type": "Name", "id": "y"},
            "value": {
                "ast_type": "Attribute",
                "attr": "counter",
                "value": {"ast_type": "Name", "id": "self"},
                "variable_reads": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 8, "source_id": 0},
                        "name": "counter",
                    }
                ],
            },
        },
        {
            "ast_type": "AugAssign",
            "op": {"ast_type": "Add"},
            "target": {
                "ast_type": "Attribute",
                "attr": "counter",
                "value": {"ast_type": "Name", "id": "lib1"},
                "variable_reads": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 29, "source_id": 1},
                        "name": "counter",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 29, "source_id": 1},
                        "name": "counter",
                    }
                ],
            },
            "value": {"ast_type": "Int", "value": 1},
        },
    ]

    assert baz["name"] == "baz"
    assert baz["body"] == [
        {
            "ast_type": "Expr",
            "value": {
                "args": [],
                "ast_type": "Call",
                "func": {
                    "ast_type": "Attribute",
                    "attr": "bar",
                    "value": {"ast_type": "Name", "id": "self"},
                    "variable_reads": [
                        {
                            "access_path": [],
                            "decl_node": {"node_id": 29, "source_id": 1},
                            "name": "counter",
                        },
                        {
                            "access_path": [],
                            "decl_node": {"node_id": 8, "source_id": 0},
                            "name": "counter",
                        },
                    ],
                    "variable_writes": [
                        {
                            "access_path": [],
                            "decl_node": {"node_id": 29, "source_id": 1},
                            "name": "counter",
                        }
                    ],
                },
                "keywords": [],
            },
        },
        {
            "ast_type": "AugAssign",
            "op": {"ast_type": "Add"},
            "target": {
                "ast_type": "Attribute",
                "attr": "counter",
                "value": {"ast_type": "Name", "id": "self"},
                "variable_reads": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 8, "source_id": 0},
                        "name": "counter",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 8, "source_id": 0},
                        "name": "counter",
                    }
                ],
            },
            "value": {"ast_type": "Int", "value": 1},
        },
    ]

    assert qux["name"] == "qux"
    assert qux["body"] == [
        {
            "ast_type": "Assign",
            "target": {
                "ast_type": "Attribute",
                "attr": "bars",
                "value": {"ast_type": "Name", "id": "lib1"},
                "variable_reads": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 34, "source_id": 1},
                        "name": "bars",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": [],
                        "decl_node": {"node_id": 34, "source_id": 1},
                        "name": "bars",
                    }
                ],
            },
            "value": {"ast_type": "List", "elements": []},
        },
        {
            "ast_type": "Assign",
            "target": {
                "ast_type": "Subscript",
                "slice": {"ast_type": "Int", "value": 0},
                "value": {
                    "ast_type": "Attribute",
                    "attr": "bars",
                    "value": {"ast_type": "Name", "id": "lib1"},
                    "variable_reads": [
                        {
                            "access_path": [],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        }
                    ],
                },
                "variable_reads": [
                    {
                        "access_path": ["$subscript_access"],
                        "decl_node": {"node_id": 34, "source_id": 1},
                        "name": "bars",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": ["$subscript_access"],
                        "decl_node": {"node_id": 34, "source_id": 1},
                        "name": "bars",
                    }
                ],
            },
            "value": {
                "args": [
                    {
                        "ast_type": "Attribute",
                        "attr": "Bar",
                        "value": {"ast_type": "Name", "id": "lib1"},
                    }
                ],
                "ast_type": "Call",
                "func": {"ast_type": "Name", "id": "empty"},
                "keywords": [],
            },
        },
        {
            "ast_type": "Assign",
            "target": {
                "ast_type": "Attribute",
                "attr": "items",
                "value": {
                    "ast_type": "Subscript",
                    "slice": {"ast_type": "Int", "value": 1},
                    "value": {
                        "ast_type": "Attribute",
                        "attr": "bars",
                        "value": {"ast_type": "Name", "id": "lib1"},
                        "variable_reads": [
                            {
                                "access_path": [],
                                "decl_node": {"node_id": 34, "source_id": 1},
                                "name": "bars",
                            }
                        ],
                    },
                    "variable_reads": [
                        {
                            "access_path": ["$subscript_access"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        }
                    ],
                },
                "variable_reads": [
                    {
                        "access_path": ["$subscript_access", "items"],
                        "decl_node": {"node_id": 34, "source_id": 1},
                        "name": "bars",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": ["$subscript_access", "items"],
                        "decl_node": {"node_id": 34, "source_id": 1},
                        "name": "bars",
                    }
                ],
            },
            "value": {
                "args": [
                    {
                        "ast_type": "Subscript",
                        "slice": {"ast_type": "Int", "value": 2},
                        "value": {
                            "ast_type": "Attribute",
                            "attr": "Foo",
                            "value": {"ast_type": "Name", "id": "lib1"},
                        },
                    }
                ],
                "ast_type": "Call",
                "func": {"ast_type": "Name", "id": "empty"},
                "keywords": [],
            },
        },
        {
            "ast_type": "Assign",
            "target": {
                "ast_type": "Attribute",
                "attr": "a",
                "value": {
                    "ast_type": "Subscript",
                    "slice": {"ast_type": "Int", "value": 0},
                    "value": {
                        "ast_type": "Attribute",
                        "attr": "items",
                        "value": {
                            "ast_type": "Subscript",
                            "slice": {"ast_type": "Int", "value": 1},
                            "value": {
                                "ast_type": "Attribute",
                                "attr": "bars",
                                "value": {"ast_type": "Name", "id": "lib1"},
                                "variable_reads": [
                                    {
                                        "access_path": [],
                                        "decl_node": {"node_id": 34, "source_id": 1},
                                        "name": "bars",
                                    }
                                ],
                            },
                            "variable_reads": [
                                {
                                    "access_path": ["$subscript_access"],
                                    "decl_node": {"node_id": 34, "source_id": 1},
                                    "name": "bars",
                                }
                            ],
                        },
                        "variable_reads": [
                            {
                                "access_path": ["$subscript_access", "items"],
                                "decl_node": {"node_id": 34, "source_id": 1},
                                "name": "bars",
                            }
                        ],
                    },
                    "variable_reads": [
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        }
                    ],
                },
                "variable_reads": [
                    {
                        "access_path": ["$subscript_access", "items", "$subscript_access", "a"],
                        "decl_node": {"node_id": 34, "source_id": 1},
                        "name": "bars",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": ["$subscript_access", "items", "$subscript_access", "a"],
                        "decl_node": {"node_id": 34, "source_id": 1},
                        "name": "bars",
                    }
                ],
            },
            "value": {"ast_type": "Int", "value": 1},
        },
        {
            "ast_type": "Assign",
            "target": {
                "ast_type": "Attribute",
                "attr": "c",
                "value": {
                    "ast_type": "Subscript",
                    "slice": {"ast_type": "Int", "value": 1},
                    "value": {
                        "ast_type": "Attribute",
                        "attr": "items",
                        "value": {
                            "ast_type": "Subscript",
                            "slice": {"ast_type": "Int", "value": 0},
                            "value": {
                                "ast_type": "Attribute",
                                "attr": "bars",
                                "value": {"ast_type": "Name", "id": "lib1"},
                                "variable_reads": [
                                    {
                                        "access_path": [],
                                        "decl_node": {"node_id": 34, "source_id": 1},
                                        "name": "bars",
                                    }
                                ],
                            },
                            "variable_reads": [
                                {
                                    "access_path": ["$subscript_access"],
                                    "decl_node": {"node_id": 34, "source_id": 1},
                                    "name": "bars",
                                }
                            ],
                        },
                        "variable_reads": [
                            {
                                "access_path": ["$subscript_access", "items"],
                                "decl_node": {"node_id": 34, "source_id": 1},
                                "name": "bars",
                            }
                        ],
                    },
                    "variable_reads": [
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        }
                    ],
                },
                "variable_reads": [
                    {
                        "access_path": ["$subscript_access", "items", "$subscript_access", "c"],
                        "decl_node": {"node_id": 34, "source_id": 1},
                        "name": "bars",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": ["$subscript_access", "items", "$subscript_access", "c"],
                        "decl_node": {"node_id": 34, "source_id": 1},
                        "name": "bars",
                    }
                ],
            },
            "value": {"ast_type": "Decimal", "value": "10.0"},
        },
    ]

    assert qux2["name"] == "qux2"
    assert qux2["body"] == [
        {
            "ast_type": "Expr",
            "value": {
                "args": [],
                "ast_type": "Call",
                "func": {
                    "ast_type": "Attribute",
                    "attr": "qux",
                    "value": {"ast_type": "Name", "id": "self"},
                    "variable_reads": [
                        {
                            "access_path": [],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                        {
                            "access_path": ["$subscript_access"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access", "a"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access", "c"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                    ],
                    "variable_writes": [
                        {
                            "access_path": [],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                        {
                            "access_path": ["$subscript_access"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access", "a"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access", "c"],
                            "decl_node": {"node_id": 34, "source_id": 1},
                            "name": "bars",
                        },
                    ],
                },
                "keywords": [],
            },
        }
    ]



import itertools
import operator
import random

import pytest

from vyper import compile_code
from vyper.exceptions import (
    InvalidOperation,
    OverflowException,
    StaticAssertionException,
    TypeMismatch,
    ZeroDivisionException,
)
from vyper.semantics.types import IntegerT
from vyper.utils import evm_div, evm_mod

types = sorted(IntegerT.signeds())


@pytest.mark.parametrize("typ", types)
def test_exponent_base_zero(get_contract, tx_failed, typ):
    code = f"""
@external
def foo(x: {typ}) -> {typ}:
    return 0 ** x
    """
    lo, hi = typ.ast_bounds

    c = get_contract(code)
    assert c.foo(0) == 1
    assert c.foo(1) == 0
    assert c.foo(hi) == 0

    with tx_failed():
        c.foo(-1)
    with tx_failed():
        c.foo(lo)  # note: lo < 0


@pytest.mark.parametrize("typ", types)
def test_exponent_base_one(get_contract, tx_failed, typ):
    code = f"""
@external
def foo(x: {typ}) -> {typ}:
    return 1 ** x
    """
    lo, hi = typ.ast_bounds

    c = get_contract(code)
    assert c.foo(0) == 1
    assert c.foo(1) == 1
    assert c.foo(hi) == 1

    with tx_failed():
        c.foo(-1)
    with tx_failed():
        c.foo(lo)


def test_exponent_base_minus_one(get_contract):
    # #2986
    # NOTE!!! python order of precedence means -1 ** x != (-1) ** x
    code = """
@external
def foo(x: int256) -> int256:
    y: int256 = (-1) ** x
    return y
    """
    c = get_contract(code)
    for x in range(5):
        assert c.foo(x) == (-1) ** x


# TODO: make this test pass
@pytest.mark.parametrize("base", (0, 1))
def test_exponent_negative_power_runtime_check(get_contract, tx_failed, base, experimental_codegen):
    # #2985
    code = f"""
@external
def bar(negative:int16) -> int16:
    x: int16 = negative
    return {base} ** x
    """
    c = get_contract(code)
    # known bug: 2985
    with tx_failed():
        c.bar(-2)


@pytest.mark.parametrize("base", (0, 1))
def test_exponent_negative_power_compile_time_check(
    get_contract, tx_failed, base, experimental_codegen
):
    # #2985
    code = f"""
@external
def bar() -> int16:
    x: int16 = -2
    return {base} ** x
    """
    if not experimental_codegen:
        return
    with pytest.raises(StaticAssertionException):
        get_contract(code)


def test_exponent_min_int16(get_contract):
    # #2987
    code = """
@external
def foo() -> int16:
    x: int16 = -8
    y: int16 = x ** 5
    return y
    """
    c = get_contract(code)
    assert c.foo() == -(2**15)


@pytest.mark.parametrize("base,power", itertools.product((-2, -1, 0, 1, 2), (0, 1)))
def test_exponent_power_zero_one(get_contract, base, power):
    # #2989
    code = f"""
@external
def foo() -> int256:
    x: int256 = {base}
    return x ** {power}
    """
    c = get_contract(code)
    assert c.foo() == base**power


@pytest.mark.parametrize("typ", types)
def test_exponent(get_contract, tx_failed, typ):
    code = f"""
@external
def foo(x: {typ}) -> {typ}:
    return 4 ** x
    """
    lo, hi = typ.ast_bounds

    c = get_contract(code)

    test_cases = [0, 1, 3, 4, 126, 127, -1, lo, hi]
    for x in test_cases:
        if x * 2 >= typ.bits or x < 0:  # out of bounds
            with tx_failed():
                c.foo(x)
        else:
            assert c.foo(x) == 4**x


@pytest.mark.parametrize("typ", types)
def test_negative_nums(get_contract, typ):
    negative_nums_code = f"""
@external
def negative_one() -> {typ}:
    return -1

@external
def negative_three() -> {typ}:
    return -(1+2)

@external
def negative_four() -> {typ}:
    a: {typ} = 2
    return -(a+2)
    """

    c = get_contract(negative_nums_code)
    assert c.negative_one() == -1
    assert c.negative_three() == -3
    assert c.negative_four() == -4


@pytest.mark.parametrize("typ", types)
def test_num_bound(tx_failed, get_contract, typ):
    lo, hi = typ.ast_bounds

    num_bound_code = f"""
@external
def _num(x: {typ}) -> {typ}:
    return x

@external
def _num_add(x: {typ}, y: {typ}) -> {typ}:
    return x + y

@external
def _num_sub(x: {typ}, y: {typ}) -> {typ}:
    return x - y

@external
def _num_add3(x: {typ}, y: {typ}, z: {typ}) -> {typ}:
    return x + y + z

@external
def _num_max() -> {typ}:
    return {hi}

@external
def _num_min() -> {typ}:
    return {lo}
    """

    c = get_contract(num_bound_code)

    assert c._num_add(hi, 0) == hi
    assert c._num_sub(lo, 0) == lo
    assert c._num_add(hi - 1, 1) == hi
    assert c._num_sub(lo + 1, 1) == lo
    with tx_failed():
        c._num_add(hi, 1)
    with tx_failed():
        c._num_sub(lo, 1)
    with tx_failed():
        c._num_add(hi - 1, 2)
    with tx_failed():
        c._num_sub(lo + 1, 2)
    assert c._num_max() == hi
    assert c._num_min() == lo

    with tx_failed():
        c._num_add3(hi, 1, -1)
    assert c._num_add3(hi, -1, 1) == hi - 1 + 1
    with tx_failed():
        c._num_add3(lo, -1, 1)
    assert c._num_add3(lo, 1, -1) == lo + 1 - 1


@pytest.mark.parametrize("typ", types)
def test_overflow_out_of_range(get_contract, typ):
    code = f"""
@external
def num_sub() -> {typ}:
    return 1-2**{typ.bits}
    """

    exc = OverflowException if typ.bits == 256 else TypeMismatch
    with pytest.raises(exc):
        compile_code(code)


ARITHMETIC_OPS = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "//": evm_div,
    "%": evm_mod,
}


@pytest.mark.parametrize("op", sorted(ARITHMETIC_OPS.keys()))
@pytest.mark.parametrize("typ", types)
@pytest.mark.fuzzing
def test_arithmetic_thorough(get_contract, tx_failed, op, typ):
    # both variables
    code_1 = f"""
@external
def foo(x: {typ}, y: {typ}) -> {typ}:
    return x {op} y
    """
    # right is literal
    code_2_template = """
@external
def foo(x: {typ}) -> {typ}:
    return x {op} {y}
    """
    # left is literal
    code_3_template = """
@external
def foo(y: {typ}) -> {typ}:
    return {x} {op} y
    """
    # both literals
    code_4_template = """
@external
def foo() -> {typ}:
    return {x} {op} {y}
    """
    lo, hi = typ.ast_bounds

    fns = {"+": operator.add, "-": operator.sub, "*": operator.mul, "//": evm_div, "%": evm_mod}
    fn = fns[op]

    c = get_contract(code_1)

    # TODO refactor to use fixtures
    special_cases = [
        lo,
        lo + 1,
        lo // 2,
        lo // 2 - 1,
        lo // 2 + 1,
        -3,
        -2,
        -1,
        0,
        1,
        2,
        3,
        hi // 2 - 1,
        hi // 2,
        hi // 2 + 1,
        hi - 1,
        hi,
    ]
    xs = special_cases.copy()
    ys = special_cases.copy()

    # note: (including special cases, roughly 8k cases total generated)

    NUM_CASES = 5
    # poor man's fuzzing - hypothesis doesn't make it easy
    # with the parametrized strategy
    xs += [random.randrange(lo, hi) for _ in range(NUM_CASES)]
    ys += [random.randrange(lo, hi) for _ in range(NUM_CASES)]

    # edge cases that are tricky to reason about and MUST be tested
    assert lo in xs and -1 in ys

    for x, y in itertools.product(xs, ys):
        expected = fn(x, y)
        in_bounds = lo <= expected <= hi

        # safediv and safemod disallow divisor == 0
        div_by_zero = y == 0 and op in ("//", "%")

        ok = in_bounds and not div_by_zero

        code_2 = code_2_template.format(typ=typ, op=op, y=y)
        code_3 = code_3_template.format(typ=typ, op=op, x=x)
        code_4 = code_4_template.format(typ=typ, op=op, x=x, y=y)

        if ok:
            assert c.foo(x, y) == expected
            assert get_contract(code_2).foo(x) == expected
            assert get_contract(code_3).foo(y) == expected
            assert get_contract(code_4).foo() == expected
        elif div_by_zero:
            with tx_failed():
                c.foo(x, y)
            with pytest.raises(ZeroDivisionException):
                compile_code(code_2)
            with tx_failed():
                get_contract(code_3).foo(y)
            with pytest.raises(ZeroDivisionException):
                compile_code(code_4)
        else:
            with tx_failed():
                c.foo(x, y)
            with tx_failed():
                get_contract(code_2).foo(x)
            with tx_failed():
                get_contract(code_3).foo(y)
            with pytest.raises((TypeMismatch, OverflowException)):
                compile_code(code_4)


COMPARISON_OPS = {
    "==": operator.eq,
    "!=": operator.ne,
    ">": operator.gt,
    ">=": operator.ge,
    "<": operator.lt,
    "<=": operator.le,
}


@pytest.mark.parametrize("op", sorted(COMPARISON_OPS.keys()))
@pytest.mark.parametrize("typ", types)
@pytest.mark.fuzzing
def test_comparators(get_contract, op, typ):
    code_1 = f"""
@external
def foo(x: {typ}, y: {typ}) -> bool:
    return x {op} y
    """

    lo, hi = typ.ast_bounds

    fn = COMPARISON_OPS[op]
    c = get_contract(code_1)

    # note: constant folding is tested in tests/unit/ast/nodes
    special_cases = [
        lo,
        lo + 1,
        lo // 2,
        lo // 2 - 1,
        lo // 2 + 1,
        -3,
        -2,
        -1,
        0,
        1,
        2,
        3,
        hi // 2 - 1,
        hi // 2,
        hi // 2 + 1,
        hi - 1,
        hi,
    ]

    xs = special_cases.copy()
    ys = special_cases.copy()

    for x, y in itertools.product(xs, ys):
        expected = fn(x, y)
        assert c.foo(x, y) is expected


@pytest.mark.parametrize("typ", types)
def test_negation(get_contract, tx_failed, typ):
    code = f"""
@external
def foo(a: {typ}) -> {typ}:
    return -a
    """

    lo, hi = typ.ast_bounds

    c = get_contract(code)

    assert c.foo(hi) == lo + 1
    assert c.foo(-1) == 1
    assert c.foo(1) == -1
    assert c.foo(0) == 0
    assert c.foo(2) == -2
    assert c.foo(-2) == 2

    with tx_failed():
        c.foo(lo)


@pytest.mark.parametrize("typ", types)
@pytest.mark.parametrize("op", ["/"])
def test_invalid_ops(get_contract, assert_compile_failed, typ, op):
    code = f"""
@external
def foo(x: {typ}, y: {typ}) -> {typ}:
    return x {op} y
    """
    assert_compile_failed(lambda: get_contract(code), InvalidOperation)


@pytest.mark.parametrize("typ", types)
@pytest.mark.parametrize("op", ["not"])
def test_invalid_unary_ops(typ, op):
    code = f"""
@external
def foo(a: {typ}) -> {typ}:
    return {op} a
    """
    with pytest.raises(InvalidOperation):
        compile_code(code)


def test_binop_nested_intermediate_underflow():
    code = """
@external
def foo():
    a: int256 = -2**255 * 2 - 10 + 100
    """
    with pytest.raises(TypeMismatch):
        compile_code(code)


def test_invalid_div():
    code = """
@external
def foo():
    a: int256 = -5 / 9
    """
    with pytest.raises(InvalidOperation) as e:
        compile_code(code)

    assert e.value._hint == "did you mean `-5 // 9`?"



import pytest

from vyper import ast as vy_ast
from vyper import semantics
from vyper.exceptions import InvalidLiteral

code_invalid_checksum = [
    """
foo: constant(address) = 0x6b175474e89094c44da98b954eedeac495271d0F
    """,
    """
foo: constant(address[1]) = [0x6b175474e89094c44da98b954eedeac495271d0F]
    """,
    """
@external
def foo():
    bar: address = 0x6b175474e89094c44da98b954eedeac495271d0F
    """,
    """
@external
def foo():
    bar: address[1] = [0x6b175474e89094c44da98b954eedeac495271d0F]
    """,
    """
@external
def foo():
    for i: address in [0x6b175474e89094c44da98b954eedeac495271d0F]:
        pass
    """,
    """
foo: constant(bytes20) = 0x6b175474e89094c44da98b954eedeac495271d0F
    """,
    """
foo: constant(bytes4) = 0x12_34_56
    """,
]


@pytest.mark.parametrize("code", code_invalid_checksum)
def test_invalid_checksum(code, dummy_input_bundle):
    with pytest.raises(InvalidLiteral):
        vyper_module = vy_ast.parse_to_ast(code)
        semantics.analyze_module(vyper_module, dummy_input_bundle)



from vyper import ast as vy_ast


def test_output_class():
    old_node = vy_ast.parse_to_ast("foo = 42")
    new_node = vy_ast.Int.from_node(old_node, value=666)

    assert isinstance(new_node, vy_ast.Int)


def test_source():
    old_node = vy_ast.parse_to_ast("foo = 42")
    new_node = vy_ast.Int.from_node(old_node, value=666)

    assert old_node.src == new_node.src
    assert old_node.node_source_code == new_node.node_source_code


def test_kwargs():
    old_node = vy_ast.parse_to_ast("42").body[0].value
    new_node = vy_ast.Int.from_node(old_node, value=666)

    assert old_node.value == 42
    assert new_node.value == 666


def test_new_node_has_no_parent():
    old_node = vy_ast.parse_to_ast("foo = 42")
    new_node = vy_ast.Int.from_node(old_node, value=666)

    assert new_node._parent is None
    assert new_node._depth == 0



from vyper.exceptions import CompilerPanic
from vyper.utils import OrderedSet
from vyper.venom.analysis.cfg import CFGAnalysis
from vyper.venom.basicblock import IRBasicBlock, IRLabel
from vyper.venom.passes.base_pass import IRPass


class SimplifyCFGPass(IRPass):
    visited: OrderedSet

    def _merge_blocks(self, a: IRBasicBlock, b: IRBasicBlock):
        a.instructions.pop()
        for inst in b.instructions:
            assert inst.opcode != "phi", "Not implemented yet"
            if inst.opcode == "phi":
                a.instructions.insert(0, inst)
            else:
                inst.parent = a
                a.instructions.append(inst)

        # Update CFG
        a.cfg_out = b.cfg_out
        if len(b.cfg_out) > 0:
            next_bb = b.cfg_out.first()
            next_bb.remove_cfg_in(b)
            next_bb.add_cfg_in(a)

            for inst in next_bb.instructions:
                if inst.opcode != "phi":
                    break
                inst.operands[inst.operands.index(b.label)] = a.label

        self.function.remove_basic_block(b)

    def _merge_jump(self, a: IRBasicBlock, b: IRBasicBlock):
        next_bb = b.cfg_out.first()
        jump_inst = a.instructions[-1]
        assert b.label in jump_inst.operands, f"{b.label} {jump_inst.operands}"
        jump_inst.operands[jump_inst.operands.index(b.label)] = next_bb.label

        # Update CFG
        a.remove_cfg_out(b)
        a.add_cfg_out(next_bb)
        next_bb.remove_cfg_in(b)
        next_bb.add_cfg_in(a)

        self.function.remove_basic_block(b)

    def _collapse_chained_blocks_r(self, bb: IRBasicBlock):
        """
        DFS into the cfg and collapse blocks with a single predecessor to the predecessor
        """
        if len(bb.cfg_out) == 1:
            next_bb = bb.cfg_out.first()
            if len(next_bb.cfg_in) == 1:
                self._merge_blocks(bb, next_bb)
                self._collapse_chained_blocks_r(bb)
                return
        elif len(bb.cfg_out) == 2:
            bb_out = bb.cfg_out.copy()
            for next_bb in bb_out:
                if (
                    len(next_bb.cfg_in) == 1
                    and len(next_bb.cfg_out) == 1
                    and len(next_bb.instructions) == 1
                ):
                    self._merge_jump(bb, next_bb)
                    self._collapse_chained_blocks_r(bb)
                    return

        if bb in self.visited:
            return
        self.visited.add(bb)

        for bb_out in bb.cfg_out:
            self._collapse_chained_blocks_r(bb_out)

    def _collapse_chained_blocks(self, entry: IRBasicBlock):
        """
        Collapse blocks with a single predecessor to their predecessor
        """
        self.visited = OrderedSet()
        self._collapse_chained_blocks_r(entry)

    def _optimize_empty_basicblocks(self) -> int:
        """
        Remove empty basic blocks.
        """
        fn = self.function
        worklist = list(fn.get_basic_blocks())
        i = count = 0
        while i < len(worklist):
            bb = worklist[i]
            i += 1

            if len(bb.instructions) > 0:
                continue

            next_bb = worklist[i]

            replaced_label = bb.label
            replacement_label = next_bb.label

            # Try to preserve symbol labels
            if replaced_label.is_symbol:
                replaced_label, replacement_label = replacement_label, replaced_label
                next_bb.label = replacement_label

            for bb2 in fn.get_basic_blocks():
                for inst in bb2.instructions:
                    for op in inst.operands:
                        if isinstance(op, IRLabel) and op.value == replaced_label.value:
                            op.value = replacement_label.value

            fn.remove_basic_block(bb)
            i -= 1
            count += 1

        return count

    def run_pass(self):
        fn = self.function
        entry = fn.entry

        for _ in range(fn.num_basic_blocks):
            changes = self._optimize_empty_basicblocks()
            changes += fn.remove_unreachable_blocks()
            if changes == 0:
                break
        else:
            raise CompilerPanic("Too many iterations removing empty basic blocks")

        self.analyses_cache.force_analysis(CFGAnalysis)

        for _ in range(fn.num_basic_blocks):  # essentially `while True`
            self._collapse_chained_blocks(entry)
            if fn.remove_unreachable_blocks() == 0:
                break
        else:
            raise CompilerPanic("Too many iterations collapsing chained blocks")

        self.analyses_cache.invalidate_analysis(CFGAnalysis)



import random

import pytest

from vyper.compiler import compile_code

valid_list = [
    """
@external
@view
def foo() -> bytes32:
    return blobhash(0)
    """,
    """
@external
@view
def foo() -> bytes32:
    a: bytes32 = 0x0000000000000000000000000000000000000000000000000000000000000005
    a = blobhash(2)
    return a
    """,
    """
@external
@view
def foo() -> bytes32:
    a: bytes32 = blobhash(0)
    assert a != empty(bytes32)
    return a
    """,
    """
@external
@view
def foo() -> bytes32:
    a: bytes32 = blobhash(1337)
    assert a == empty(bytes32)
    return a
    """,
]


@pytest.mark.requires_evm_version("cancun")
@pytest.mark.parametrize("good_code", valid_list)
def test_blobhash_success(good_code):
    assert compile_code(good_code) is not None
    out = compile_code(good_code, output_formats=["opcodes_runtime"])
    assembly = out["opcodes_runtime"].split(" ")
    assert "BLOBHASH" in assembly


@pytest.mark.requires_evm_version("cancun")
def test_get_blobhashes(env, get_contract, tx_failed):
    code = """
@external
def get_blobhash(i: uint256) -> bytes32:
    return blobhash(i)
"""
    c = get_contract(code)

    # mock the evm blobhash attribute
    env.blob_hashes = [random.randbytes(32) for _ in range(6)]

    for i in range(6):
        assert c.get_blobhash(i) == env.blob_hashes[i]

    assert c.get_blobhash(len(env.blob_hashes)) == b"\0" * 32



import pytest

from vyper import ast as vy_ast
from vyper.exceptions import SyntaxException


def test_binary_becomes_bytes():
    expected = vy_ast.parse_to_ast(
        """
def x():
    foo: Bytes[1] = b'\x01'
    """
    )
    mutated = vy_ast.parse_to_ast(
        """
def x():
    foo: Bytes[1] = 0b00000001
    """
    )

    assert expected == mutated


def test_binary_length():
    with pytest.raises(SyntaxException):
        vy_ast.parse_to_ast("foo: Bytes[1] = 0b01")



import pytest

from vyper import compile_code
from vyper.exceptions import InvalidLiteral, InvalidType

fail_list = [
    (
        """
@external
def foo():
    a: Bytes[4] = method_id("bar ()")
    """,
        InvalidLiteral,
    ),
    (
        """
FOO: constant(Bytes[4]) = method_id(1)
    """,
        InvalidType,
    ),
    (
        """
FOO: constant(Bytes[4]) = method_id("bar ()")
    """,
        InvalidLiteral,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_method_id_fail(bad_code, exc):
    with pytest.raises(exc):
        compile_code(bad_code)


valid_list = [
    """
FOO: constant(String[5]) = "foo()"
BAR: constant(Bytes[4]) = method_id(FOO)

@external
def foo(a: Bytes[4] = BAR):
    pass
    """
]


@pytest.mark.parametrize("code", valid_list)
def test_method_id_pass(code):
    assert compile_code(code) is not None



import math

import hypothesis
import pytest

from vyper.utils import SizeLimits


@pytest.fixture(scope="module")
def isqrt_contract(get_contract):
    code = """
@external
def test(a: uint256) -> uint256:
    return isqrt(a)
    """
    c = get_contract(code)
    return c


def test_isqrt_literal(get_contract):
    val = 2
    code = f"""
@external
def test() -> uint256:
    return isqrt({val})
    """
    c = get_contract(code)
    assert c.test() == math.isqrt(val)


def test_isqrt_variable(get_contract):
    code = """
@external
def test(a: uint256) -> uint256:
    return isqrt(a)
    """

    c = get_contract(code)

    val = 3333
    assert c.test(val) == math.isqrt(val)

    val = 10**17
    assert c.test(val) == math.isqrt(val)
    assert c.test(0) == 0


def test_isqrt_internal_variable(get_contract):
    val = 44001
    code = f"""
@external
def test2() -> uint256:
    a: uint256 = {val}
    return isqrt(a)
    """
    c = get_contract(code)
    assert c.test2() == math.isqrt(val)


def test_isqrt_storage(get_contract):
    code = """
s_var: uint256

@external
def test(a: uint256) -> uint256:
    self.s_var = a + 1
    return isqrt(self.s_var)
    """

    c = get_contract(code)
    val = 1221
    assert c.test(val) == math.isqrt(val + 1)
    val = 10001
    assert c.test(val) == math.isqrt(val + 1)


def test_isqrt_storage_internal_variable(get_contract):
    val = 44444
    code = f"""
s_var: uint256

@external
def test2() -> uint256:
    self.s_var = {val}
    return isqrt(self.s_var)
    """
    c = get_contract(code)
    assert c.test2() == math.isqrt(val)


def test_isqrt_inline_memory_correct(get_contract):
    code = """
@external
def test(a: uint256) -> (uint256, uint256, uint256, uint256, uint256, String[100]):
    x: uint256 = 1
    y: uint256 = 2
    z: uint256 = 3
    e: uint256 = isqrt(a)
    f: String[100] = 'hello world'
    return a, x, y, z, e, f
    """

    c = get_contract(code)

    val = 21
    assert c.test(val) == (val, 1, 2, 3, math.isqrt(val), "hello world")


@pytest.mark.fuzzing
@hypothesis.given(
    value=hypothesis.strategies.integers(min_value=0, max_value=SizeLimits.MAX_UINT256)
)
@hypothesis.example(SizeLimits.MAX_UINT256)
@hypothesis.example(0)
@hypothesis.example(1)
# the following examples demonstrate correct rounding mode
# for an edge case in the babylonian method - the operand is
# a perfect square - 1
@hypothesis.example(2704)
@hypothesis.example(110889)
@hypothesis.example(32239684)
def test_isqrt_valid_range(isqrt_contract, value):
    vyper_isqrt = isqrt_contract.test(value)
    actual_isqrt = math.isqrt(value)
    assert vyper_isqrt == actual_isqrt

    # check if sqrt limits are correct
    next = vyper_isqrt + 1
    assert vyper_isqrt * vyper_isqrt <= value
    assert next * next > value



import pytest

from vyper import compile_code
from vyper.exceptions import InvalidType

# CMC 2023-12-31 this could probably go in builtins/folding/
fail_list = [
    (
        """
FOO: constant(address) = epsilon(address)
    """,
        InvalidType,
    )
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_block_fail(bad_code, exc):
    with pytest.raises(exc):
        compile_code(bad_code)



import pytest

from vyper import compiler
from vyper.exceptions import FunctionDeclarationException

fail_list = [
    """
x: int128
@external
@const
def foo() -> int128:
    pass
    """,
    """
x: int128
@external
@monkeydoodledoo
def foo() -> int128:
    pass
    """,
    """
def foo() -> int128:
    q: int128 = 111
    return q
    """,
    """
q: int128
def foo() -> int128:
    return self.q
    """,
    """
@external
def test_func() -> int128:
    return (1, 2)
    """,
    """
@deploy
def __init__(a: int128 = 12):
    pass
    """,
    """
@deploy
def __init__() -> uint256:
    return 1
    """,
    """
@deploy
def __init__() -> bool:
    pass
    """,
    """
a: immutable(uint256)

@internal
def __init__():
    a = 1
    """,
    """
a: immutable(uint256)

@deploy
@pure
def __init__():
    a = 1
    """,
    """
a: immutable(uint256)

@deploy
@view
def __init__():
    a = 1
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_function_declaration_exception(bad_code):
    with pytest.raises(FunctionDeclarationException):
        compiler.compile_code(bad_code)



from decimal import ROUND_FLOOR, Decimal

import hypothesis
import pytest

from tests.utils import decimal_to_int
from vyper.utils import SizeLimits

DECIMAL_PLACES = 10
DECIMAL_RANGE = [Decimal("0." + "0" * d + "2") for d in range(0, DECIMAL_PLACES)]


def decimal_truncate(val, decimal_places=DECIMAL_PLACES, rounding=ROUND_FLOOR):
    q = "0"
    if decimal_places != 0:
        q += "." + "0" * decimal_places

    return val.quantize(Decimal(q), rounding=rounding)


def decimal_sqrt(val):
    return decimal_to_int(decimal_truncate(val.sqrt()))


def test_sqrt_literal(get_contract):
    code = """
@external
def test() -> decimal:
    return sqrt(2.0)
    """
    c = get_contract(code)
    assert c.test() == decimal_sqrt(Decimal("2"))


# TODO: use parametrization here
def test_sqrt_variable(get_contract):
    code = """
@external
def test(a: decimal) -> decimal:
    return sqrt(a)

@external
def test2() -> decimal:
    a: decimal = 44.001
    return sqrt(a)
    """

    c = get_contract(code)

    val = Decimal("33.33")
    assert c.test(decimal_to_int(val)) == decimal_sqrt(val)

    val = Decimal("0.1")
    assert c.test(decimal_to_int(val)) == decimal_sqrt(val)

    assert c.test(decimal_to_int("0.0")) == decimal_to_int("0.0")
    assert c.test2() == decimal_sqrt(Decimal("44.001"))


def test_sqrt_storage(get_contract):
    code = """
s_var: decimal

@external
def test(a: decimal) -> decimal:
    self.s_var = a + 1.0
    return sqrt(self.s_var)

@external
def test2() -> decimal:
    self.s_var = 444.44
    return sqrt(self.s_var)
    """

    c = get_contract(code)
    val = Decimal("12.21")
    assert c.test(decimal_to_int(val)) == decimal_sqrt(val + 1)
    val = Decimal("100.01")
    assert c.test(decimal_to_int(val)) == decimal_sqrt(val + 1)
    assert c.test2() == decimal_sqrt(Decimal("444.44"))


def test_sqrt_inline_memory_correct(get_contract):
    code = """
@external
def test(a: decimal) -> (decimal, decimal, decimal, decimal, decimal, String[100]):
    x: decimal = 1.0
    y: decimal = 2.0
    z: decimal = 3.0
    e: decimal = sqrt(a)
    f: String[100] = 'hello world'
    return a, x, y, z, e, f
    """

    c = get_contract(code)

    val = Decimal("2.1")
    assert c.test(decimal_to_int(val)) == (
        decimal_to_int(val),
        decimal_to_int("1"),
        decimal_to_int("2"),
        decimal_to_int("3"),
        decimal_sqrt(val),
        "hello world",
    )


@pytest.mark.parametrize("value", DECIMAL_RANGE)
def test_sqrt_sub_decimal_places(value, get_contract):
    code = """
@external
def test(a: decimal) -> decimal:
    return sqrt(a)
    """

    c = get_contract(code)

    vyper_sqrt = c.test(decimal_to_int(value))
    actual_sqrt = decimal_sqrt(value)
    assert vyper_sqrt == actual_sqrt


@pytest.fixture(scope="module")
def sqrt_contract(get_contract):
    code = """
@external
def test(a: decimal) -> decimal:
    return sqrt(a)
    """
    c = get_contract(code)
    return c


@pytest.mark.parametrize("value", [Decimal(0), Decimal(SizeLimits.MAX_INT128)])
def test_sqrt_bounds(sqrt_contract, value):
    vyper_sqrt = sqrt_contract.test(decimal_to_int(value))
    actual_sqrt = decimal_sqrt(value)
    assert vyper_sqrt == actual_sqrt


@pytest.mark.fuzzing
@hypothesis.given(
    value=hypothesis.strategies.decimals(
        min_value=Decimal(0), max_value=Decimal(SizeLimits.MAX_INT128), places=DECIMAL_PLACES
    )
)
@hypothesis.example(value=Decimal(SizeLimits.MAX_INT128))
@hypothesis.example(value=Decimal(0))
def test_sqrt_valid_range(sqrt_contract, value):
    vyper_sqrt = sqrt_contract.test(decimal_to_int(value))
    actual_sqrt = decimal_sqrt(value)
    assert vyper_sqrt == actual_sqrt


@pytest.mark.fuzzing
@hypothesis.given(
    value=hypothesis.strategies.decimals(
        min_value=Decimal(SizeLimits.MIN_INT128), max_value=Decimal("-1E10"), places=DECIMAL_PLACES
    )
)
@hypothesis.example(value=Decimal(SizeLimits.MIN_INT128))
@hypothesis.example(value=Decimal("-1E10"))
def test_sqrt_invalid_range(tx_failed, sqrt_contract, value):
    with tx_failed():
        sqrt_contract.test(decimal_to_int(value))



def test_values_should_be_increasing_ints(get_contract):
    code = """
flag Action:
    BUY
    SELL
    CANCEL

@external
@view
def buy() -> Action:
    return Action.BUY

@external
@view
def sell() -> Action:
    return Action.SELL

@external
@view
def cancel() -> Action:
    return Action.CANCEL
    """
    c = get_contract(code)
    assert c.buy() == 1
    assert c.sell() == 2
    assert c.cancel() == 4


def test_flag_storage(get_contract):
    code = """
flag Actions:
    BUY
    SELL
    CANCEL

action: public(Actions)

@external
def set_and_get(a: Actions) -> Actions:
    self.action = a
    return self.action
    """
    c = get_contract(code)
    for i in range(5):
        assert c.set_and_get(i) == i
        c.set_and_get(i)
        assert c.action() == i


def test_eq_neq(get_contract):
    code = """
flag Roles:
    USER
    STAFF
    ADMIN
    MANAGER
    CEO

@external
def is_boss(a: Roles) -> bool:
    return a == Roles.CEO

@external
def is_not_boss(a: Roles) -> bool:
    return a != Roles.CEO
    """
    c = get_contract(code)

    for i in range(4):
        assert c.is_boss(2**i) is False
        assert c.is_not_boss(2**i) is True

    assert c.is_boss(2**4) is True
    assert c.is_not_boss(2**4) is False


def test_bitwise(get_contract, tx_failed):
    code = """
flag Roles:
    USER
    STAFF
    ADMIN
    MANAGER
    CEO

@external
def bor() -> Roles:
    return Roles.USER | Roles.CEO

@external
def band() -> Roles:
    c: Roles = Roles.USER | Roles.CEO
    return c & Roles.USER

@external
def bxor() -> Roles:
    c: Roles = Roles.USER | Roles.CEO
    return c ^ Roles.USER

@external
def binv() -> Roles:
    c: Roles = Roles.USER
    return ~c

@external
def bor_arg(a: Roles, b: Roles) -> Roles:
    return a | b

@external
def band_arg(a: Roles, b: Roles) -> Roles:
    return a & b

@external
def bxor_arg(a: Roles, b: Roles) -> Roles:
    return a ^ b

@external
def binv_arg(a: Roles) -> Roles:
    return ~a
    """
    c = get_contract(code)
    assert c.bor() == 17
    assert c.band() == 1
    assert c.bxor() == 16

    assert c.bor_arg(0b00001, 0b00100) == 0b00001 | 0b00100 == 0b00101
    # LHS: USER | ADMIN | CEO; RHS: USER | MANAGER | CEO
    assert c.band_arg(0b10101, 0b11001) == 0b10101 & 0b11001 == 0b10001

    assert c.bxor_arg(0b10101, 0b11001) == 0b10101 ^ 0b11001 == 0b01100

    assert c.binv_arg(0b01101) == ~0b01101 % 32 == 0b10010
    assert c.binv_arg(0b11111) == 0b00000
    assert c.binv_arg(0b00000) == 0b11111

    # LHS is out of bound
    with tx_failed():
        c.bor_arg(32, 3)
    with tx_failed():
        c.band_arg(32, 3)
    with tx_failed():
        c.bxor_arg(32, 3)
    with tx_failed():
        c.binv_arg(32)

    # RHS
    with tx_failed():
        c.bor_arg(3, 32)
    with tx_failed():
        c.band_arg(3, 32)
    with tx_failed():
        c.bxor_arg(3, 32)


def test_augassign_storage(get_contract, env, tx_failed):
    code = """
flag Roles:
    ADMIN
    MINTER

roles: public(HashMap[address, Roles])

@deploy
def __init__():
    self.roles[msg.sender] = Roles.ADMIN

@external
def addMinter(minter: address):
    assert self.roles[msg.sender] in Roles.ADMIN
    self.roles[minter] |= Roles.MINTER

@external
def revokeMinter(minter: address):
    assert self.roles[msg.sender] in Roles.ADMIN
    self.roles[minter] &= ~Roles.MINTER

@external
def flipMinter(minter: address):
    assert self.roles[msg.sender] in Roles.ADMIN
    self.roles[minter] ^= Roles.MINTER

@external
def checkMinter(minter: address):
    assert Roles.MINTER in self.roles[minter]
    """
    c = get_contract(code)

    # check admin
    admin_address = env.deployer
    minter_address = env.accounts[1]

    # add minter
    c.addMinter(minter_address)
    c.checkMinter(minter_address)

    assert c.roles(admin_address) == 0b01
    assert c.roles(minter_address) == 0b10

    # admin is not a minter
    with tx_failed():
        c.checkMinter(admin_address)

    c.addMinter(admin_address)

    # now, admin is a minter
    assert c.roles(admin_address) == 0b11
    c.checkMinter(admin_address)

    # revoke minter
    c.revokeMinter(admin_address)
    assert c.roles(admin_address) == 0b01
    with tx_failed():
        c.checkMinter(admin_address)

    # flip minter
    c.flipMinter(admin_address)
    assert c.roles(admin_address) == 0b11
    c.checkMinter(admin_address)

    # flip minter
    c.flipMinter(admin_address)
    assert c.roles(admin_address) == 0b01
    with tx_failed():
        c.checkMinter(admin_address)


def test_in_flag(get_contract):
    code = """
flag Roles:
    USER
    STAFF
    ADMIN
    MANAGER
    CEO

@external
def foo() -> bool:
    return Roles.USER in (Roles.USER | Roles.ADMIN)

@external
def bar(a: Roles) -> bool:
    return a in (Roles.USER | Roles.ADMIN)

@external
def bar2(a: Roles) -> bool:
    return a not in (Roles.USER | Roles.ADMIN)

@external
def baz(a: Roles) -> bool:
    x: Roles = Roles.USER | Roles.ADMIN | Roles.CEO
    y: Roles = x ^ (Roles.MANAGER | Roles.CEO)  # flip off CEO, flip on MANAGER
    return a in (x & y)

    """
    c = get_contract(code)
    assert c.foo() is True

    # CEO MANAGER ADMIN STAFF USER
    #   1       1     1     1    1

    assert c.bar(0b00001) is True  # Roles.USER should pass
    assert c.bar(0b00010) is False  # Roles.STAFF should fail

    assert c.bar2(0b00001) is False  # Roles.USER should fail
    assert c.bar2(0b00010) is True  # Roles.STAFF should pass

    assert c.baz(0b00001) is True  # Roles.USER should pass
    assert c.baz(0b00100) is True  # Roles.ADMIN should pass
    assert c.baz(0b01000) is False  # Roles.MANAGER should fail


def test_struct_with_flag(get_contract):
    code = """
flag Foobar:
    FOO
    BAR

struct Foo:
    a: uint256
    b: Foobar

@external
def get_flag_from_struct() -> Foobar:
    f: Foo = Foo(a=1, b=Foobar.BAR)
    return f.b
    """
    c = get_contract(code)
    assert c.get_flag_from_struct() == 2


def test_mapping_with_flag(get_contract):
    code = """
flag Foobar:
    FOO
    BAR

fb: HashMap[Foobar, uint256]

@external
def get_key(f: Foobar, i: uint256) -> uint256:
    self.fb[f] = i
    return self.fb[f]
    """
    c = get_contract(code)
    assert c.get_key(1, 777) == 777



import copy
import logging
from contextlib import contextmanager
from typing import Optional

import rlp
from cached_property import cached_property
from eth.abc import ChainAPI, ComputationAPI, VirtualMachineAPI
from eth.chains.mainnet import MainnetChain
from eth.constants import CREATE_CONTRACT_ADDRESS, GENESIS_DIFFICULTY
from eth.db.atomic import AtomicDB
from eth.exceptions import Revert, VMError
from eth.tools.builder import chain as chain_builder
from eth.vm.base import StateAPI
from eth.vm.execution_context import ExecutionContext
from eth.vm.forks.cancun.transaction_context import CancunTransactionContext
from eth.vm.message import Message
from eth_keys.datatypes import PrivateKey
from eth_typing import Address
from eth_utils import setup_DEBUG2_logging, to_canonical_address, to_checksum_address

import vyper.evm.opcodes as evm_opcodes
from tests.evm_backends.base_env import BaseEnv, EvmError, ExecutionResult, LogEntry
from vyper.utils import keccak256


class PyEvmEnv(BaseEnv):
    """EVM backend environment using the Py-EVM library."""

    invalid_opcode_error = "Invalid opcode"
    out_of_gas_error = "Out of gas"

    def __init__(
        self,
        gas_limit: int,
        account_keys: list[PrivateKey],
        tracing: bool,
        block_number: int,
        evm_version: str,
    ) -> None:
        super().__init__(gas_limit, account_keys)

        evm_opcodes.DEFAULT_EVM_VERSION = evm_version

        if tracing:
            logger = logging.getLogger("eth.vm.computation.BaseComputation")
            setup_DEBUG2_logging()
            logger.setLevel("DEBUG2")
            # from vdb import vdb
            # vdb.set_evm_opcode_debugger()

        spec = getattr(chain_builder, evm_version + "_at")(block_number)
        self._chain: ChainAPI = chain_builder.build(MainnetChain, spec).from_genesis(
            base_db=AtomicDB(),
            genesis_params={"difficulty": GENESIS_DIFFICULTY, "gas_limit": gas_limit},
        )

        self._last_computation: ComputationAPI = None
        self._blob_hashes: list[bytes] = []

    @cached_property
    def _state(self) -> StateAPI:
        return self._vm.state

    @cached_property
    def _vm(self) -> VirtualMachineAPI:
        return self._chain.get_vm()

    @property
    def _context(self) -> ExecutionContext:
        context = self._state.execution_context
        assert isinstance(context, ExecutionContext)  # help mypy
        return context

    @contextmanager
    def anchor(self):
        snapshot_id = self._state.snapshot()
        ctx = copy.copy(self._state.execution_context)
        try:
            yield
        finally:
            self._state.revert(snapshot_id)
            self._state.execution_context = ctx

    def get_balance(self, address: str) -> int:
        return self._state.get_balance(_addr(address))

    def set_balance(self, address: str, value: int):
        self._state.set_balance(_addr(address), value)

    @property
    def block_number(self) -> int:
        return self._context.block_number

    @block_number.setter
    def block_number(self, value: int):
        self._context._block_number = value

    @property
    def timestamp(self) -> int | None:
        return self._context.timestamp

    @timestamp.setter
    def timestamp(self, value: int):
        self._context._timestamp = value

    @property
    def last_result(self) -> ExecutionResult:
        result = self._last_computation
        return ExecutionResult(
            is_success=not result.is_error,
            logs=list(_parse_log_entries(result)),
            gas_refunded=result.get_gas_refund(),
            gas_used=result.get_gas_used(),
        )

    @property
    def blob_hashes(self) -> list[bytes]:
        return self._blob_hashes

    @blob_hashes.setter
    def blob_hashes(self, value: list[bytes]):
        self._blob_hashes = value

    def message_call(
        self,
        to: str,
        sender: str | None = None,
        data: bytes | str = b"",
        value: int = 0,
        gas: int | None = None,
        gas_price: int = 0,
        is_modifying: bool = True,
    ):
        if isinstance(data, str):
            data = bytes.fromhex(data.removeprefix("0x"))
        sender = _addr(sender or self.deployer)
        try:
            computation = self._state.computation_class.apply_message(
                state=self._state,
                message=Message(
                    to=_addr(to),
                    sender=sender,
                    data=data,
                    code=self.get_code(to),
                    value=value,
                    gas=self.gas_limit if gas is None else gas,
                    is_static=not is_modifying,
                ),
                transaction_context=self._make_tx_context(sender, gas_price),
            )
        except VMError as e:
            # py-evm raises when user is out-of-funds instead of returning a failed computation
            raise EvmError(*e.args) from e

        self._check_computation(computation)
        return computation.output

    def _make_tx_context(self, sender, gas_price):
        context_class = self._state.transaction_context_class
        context = context_class(origin=sender, gas_price=gas_price)
        if self._blob_hashes:
            assert isinstance(context, CancunTransactionContext)
            context._blob_versioned_hashes = self._blob_hashes
        return context

    def clear_transient_storage(self) -> None:
        try:
            self._state.clear_transient_storage()
        except AttributeError as e:
            assert e.args == ("No transient_storage has been set for this State",)

    def _check_computation(self, computation):
        self._last_computation = computation
        if computation.is_error:
            if isinstance(computation.error, Revert):
                (output,) = computation.error.args
                gas_used = computation.get_gas_used()
                self._parse_revert(output, computation.error, gas_used)

            raise EvmError(*computation.error.args) from computation.error

    def get_code(self, address: str):
        return self._state.get_code(_addr(address))

    def get_excess_blob_gas(self) -> Optional[int]:
        return self._context.excess_blob_gas

    def set_excess_blob_gas(self, param):
        self._context._excess_blob_gas = param

    def _deploy(self, code: bytes, value: int, gas: int = None) -> str:
        sender = _addr(self.deployer)
        target_address = self._generate_contract_address(sender)

        try:
            computation = self._state.computation_class.apply_create_message(
                state=self._state,
                message=Message(
                    to=CREATE_CONTRACT_ADDRESS,  # i.e., b""
                    sender=sender,
                    value=value,
                    code=code,
                    data=b"",
                    gas=gas or self.gas_limit,
                    create_address=target_address,
                ),
                transaction_context=self._make_tx_context(sender, gas_price=0),
            )
        except VMError as e:
            # py-evm raises when user is out-of-funds instead of returning a failed computation
            raise EvmError(*e.args) from e

        self._check_computation(computation)
        return "0x" + target_address.hex()

    def _generate_contract_address(self, sender: Address) -> Address:
        nonce = self._state.get_nonce(sender)
        self._state.increment_nonce(sender)
        next_account_hash = keccak256(rlp.encode([sender, nonce]))
        return to_canonical_address(next_account_hash[-20:])


def _parse_log_entries(result: ComputationAPI):
    """
    Parses the raw log entries from a computation result into a more
    usable format similar to the revm backend.
    """
    for address, topics, data in result.get_log_entries():
        topic_bytes = [t.to_bytes(32, "big") for t in topics]
        topic_ids = ["0x" + t.hex() for t in topic_bytes]
        yield LogEntry(to_checksum_address(address), topic_ids, (topic_bytes, data))


def _addr(address: str) -> Address:
    """Convert an address string to an Address object."""
    return Address(bytes.fromhex(address.removeprefix("0x")))



import pytest

from vyper import compiler
from vyper.exceptions import TypeMismatch

fail_list = [
    (
        """
@external
def foo(j: uint256) -> bool:
    s: bool = _abi_decode(j, bool, unwrap_tuple= False)
    return s
    """,
        TypeMismatch,
    ),
    (
        """
@external
def bar(j: String[32]) -> bool:
    s: bool = _abi_decode(j, bool, unwrap_tuple= False)
    return s
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_abi_decode_fail(bad_code, exc):
    with pytest.raises(exc):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo(x: Bytes[32]) -> uint256:
    return _abi_decode(x, uint256)
    """
]


@pytest.mark.parametrize("good_code", valid_list)
def test_abi_decode_success(good_code):
    assert compiler.compile_code(good_code) is not None



def test_module_event(get_contract, make_input_bundle, get_logs):
    # log from a module
    lib1 = """
event MyEvent:
    pass

@internal
def foo():
    log MyEvent()
    """
    main = """
import lib1

@external
def bar():
    lib1.foo()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)
    c.bar()
    logs = get_logs(c, "MyEvent")
    assert len(logs) == 1


def test_module_event2(get_contract, make_input_bundle, get_logs):
    # log a module event from main contract
    lib1 = """
event MyEvent:
    x: uint256
    """
    main = """
import lib1

@external
def bar():
    log lib1.MyEvent(5)
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)
    c.bar()
    (log,) = get_logs(c, "MyEvent")
    assert log.args.x == 5


def test_module_event_indexed(get_contract, make_input_bundle, get_logs):
    lib1 = """
event MyEvent:
    x: uint256
    y: indexed(uint256)

@internal
def foo():
    log MyEvent(5, 6)
    """
    main = """
import lib1

@external
def bar():
    lib1.foo()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)
    c.bar()
    (log,) = get_logs(c, "MyEvent")
    assert log.args.x == 5
    assert log.args.y == 6



import pytest

from vyper import ast as vy_ast
from vyper.semantics.namespace import get_namespace


@pytest.fixture(scope="session")
def build_node():
    """
    Yields a helper function for generating a single Vyper AST node.
    """

    def _build_node(source):
        # docstring ensures string nodes are properly generated, not turned into docstrings
        source = f"""'I am a docstring.'\n{source}"""
        ast = vy_ast.parse_to_ast(source).body[0]
        if isinstance(ast, vy_ast.Expr):
            ast = ast.value
        return ast

    yield _build_node


@pytest.fixture
def namespace():
    """
    Yields a clean `Namespace` object.
    """
    obj = get_namespace()
    obj.clear()
    yield obj
    obj.clear()



import pytest

from vyper.exceptions import SyntaxException

fail_list = [
    """
x: Bytes[1:3]
    """,
    """
b: int128[int128: address]
    """,
    """
x: int128[5]
@external
def foo():
    self.x[2:4] = 3
    """,
    """
x: int128[5]
@external
def foo():
    z = self.x[2:4]
    """,
    """
@external
def foo():
    x: int128[5]
    z = x[2:4]
    """,
    """
Transfer: event({_rom&: indexed(address)})
    """,
    """
@external
def test() -> uint256:
    for i in range(0, 4):
      return 0
    else:
      return 1
    return 1
    """,
    """
@external
def foo():
    x = y = 3
    """,
    """
@external
def foo():
    x: address = create_minimal_proxy_to(0x123456789012345678901234567890123456789)
    """,
    """
@external
def foo():
    x: Bytes[4] = raw_call(0x123456789012345678901234567890123456789, "cow", max_outsize=4)
    """,
    """
@external
def foo():
    x: address = 0x12345678901234567890123456789012345678901
    """,
    """
@external
def foo():
    x: address = 0x01234567890123456789012345678901234567890
    """,
    """
@external
def foo():
    x: address = 0x123456789012345678901234567890123456789
    """,
    """
a: internal(uint256)
    """,
    """
@external
def foo():
    x: uint256 = +1  # test UAdd ast blocked
    """,
    """
@internal
def f(a:uint256,/):  # test posonlyargs blocked
    return

@external
def g():
    self.f()
    """,
    """
@external
def foo():
    for i in range(0, 10):
        pass
    """,
    """
@external
def foo():
    for i: $$$ in range(0, 10):
        pass
    """,
    """
struct S:
    x: int128
s: S = S(x=int128, 1)
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_syntax_exception(assert_compile_failed, get_contract, bad_code):
    assert_compile_failed(lambda: get_contract(bad_code), SyntaxException)



from tests.utils import ZERO_ADDRESS


def test_send(tx_failed, get_contract, env):
    send_test = """
@external
def foo():
    send(msg.sender, self.balance + 1)

@external
def fop():
    send(msg.sender, 10)
    """
    env.set_balance(env.deployer, 1000)
    c = get_contract(send_test, value=10)
    with tx_failed():
        c.foo()
    assert env.get_balance(c.address) == 10
    c.fop()
    assert env.get_balance(c.address) == 0
    with tx_failed():
        c.fop()


def test_default_gas(get_contract, env, tx_failed):
    """
    Tests to verify that send to default function will send limited gas (2300),
    but raw_call can send more.
    """

    sender_code = """
@external
def test_send(receiver: address):
    send(receiver, 1)

@external
def test_call(receiver: address):
    raw_call(receiver, b"", gas=50000, max_outsize=0, value=1)
    """

    # default function writes variable, this requires more gas than send can pass
    receiver_code = """
last_sender: public(address)

@external
@payable
def __default__():
    self.last_sender = msg.sender
    """

    env.set_balance(env.deployer, 300000)
    sender = get_contract(sender_code, value=1)
    receiver = get_contract(receiver_code)

    with tx_failed():
        sender.test_send(receiver.address, gas=100000)

    # no value transfer happened, variable was not changed
    assert receiver.last_sender() == ZERO_ADDRESS
    assert env.get_balance(sender.address) == 1
    assert env.get_balance(receiver.address) == 0

    sender.test_call(receiver.address, gas=100000)

    # value transfer happened, variable was changed
    assert receiver.last_sender() == sender.address
    assert env.get_balance(sender.address) == 0
    assert env.get_balance(receiver.address) == 1


def test_send_gas_stipend(get_contract, env):
    """
    Tests to verify that adding gas stipend to send() will send sufficient gas
    """

    sender_code = """

@external
def test_send_stipend(receiver: address):
    send(receiver, 1, gas=50000)
    """

    # default function writes variable, this requires more gas than
    # send would pass without gas stipend
    receiver_code = """
last_sender: public(address)

@external
@payable
def __default__():
    self.last_sender = msg.sender
    """

    env.set_balance(env.deployer, 300000)
    sender = get_contract(sender_code, value=1)
    receiver = get_contract(receiver_code)

    sender.test_send_stipend(receiver.address, gas=100000)

    # value transfer happened, variable was changed
    assert receiver.last_sender() == sender.address
    assert env.get_balance(sender.address) == 0
    assert env.get_balance(receiver.address) == 1



import json
from pathlib import PurePath

import pytest

import vyper
from vyper.cli.vyper_json import (
    compile_from_input_dict,
    compile_json,
    exc_handler_to_dict,
    get_inputs,
)
from vyper.compiler import OUTPUT_FORMATS, compile_code, compile_from_file_input
from vyper.compiler.input_bundle import JSONInputBundle
from vyper.exceptions import JSONError, SyntaxException, TypeMismatch

FOO_CODE = """
import contracts.ibar as IBar

import contracts.library as library

@external
def foo(a: address) -> bool:
    return extcall IBar(a).bar(1)

@external
def baz() -> uint256:
    return self.balance + library.foo()
"""

BAR_CODE = """
import contracts.ibar as IBar

implements: IBar

@external
def bar(a: uint256) -> bool:
    return True
"""

BAR_VYI = """
@external
def bar(a: uint256) -> bool:
    ...
"""

LIBRARY_CODE = """
@internal
def foo() -> uint256:
    return block.number + 1
"""

BAD_SYNTAX_CODE = """
def bar()>:
"""

BAD_COMPILER_CODE = """
@external
def oopsie(a: uint256) -> bool:
    return 42
"""

BAR_ABI = [
    {
        "name": "bar",
        "outputs": [{"type": "bool", "name": "out"}],
        "inputs": [{"type": "uint256", "name": "a"}],
        "stateMutability": "nonpayable",
        "type": "function",
    }
]


@pytest.fixture(scope="function")
def input_json(optimize, evm_version, experimental_codegen):
    return {
        "language": "Vyper",
        "sources": {
            "contracts/foo.vy": {"content": FOO_CODE},
            "contracts/library.vy": {"content": LIBRARY_CODE},
            "contracts/bar.vy": {"content": BAR_CODE},
        },
        "interfaces": {"contracts/ibar.json": {"abi": BAR_ABI}},
        "settings": {
            "outputSelection": {"*": ["*"]},
            "optimize": optimize.name.lower(),
            "evmVersion": evm_version,
            "experimentalCodegen": experimental_codegen,
        },
    }


@pytest.fixture(scope="function")
def input_bundle(input_json):
    # CMC 2023-12-11 maybe input_json -> JSONInputBundle should be a helper
    # function in `vyper_json.py`.
    sources = get_inputs(input_json)
    return JSONInputBundle(sources, search_paths=[PurePath(".")])


# test string and dict inputs both work
def test_string_input(input_json):
    assert compile_json(input_json) == compile_json(json.dumps(input_json))


def test_bad_json():
    with pytest.raises(JSONError):
        compile_json("this probably isn't valid JSON, is it")


def test_keyerror_becomes_jsonerror(input_json):
    del input_json["sources"]
    with pytest.raises(KeyError):
        compile_from_input_dict(input_json)
    with pytest.raises(JSONError):
        compile_json(input_json)


def test_compile_json(input_json, input_bundle):
    foo_input = input_bundle.load_file("contracts/foo.vy")
    # remove venom related from output formats
    # because they require venom (experimental)
    output_formats = OUTPUT_FORMATS.copy()
    del output_formats["bb"]
    del output_formats["bb_runtime"]
    del output_formats["cfg"]
    del output_formats["cfg_runtime"]
    foo = compile_from_file_input(
        foo_input, output_formats=output_formats, input_bundle=input_bundle
    )

    library_input = input_bundle.load_file("contracts/library.vy")
    library = compile_from_file_input(
        library_input, output_formats=output_formats, input_bundle=input_bundle
    )

    bar_input = input_bundle.load_file("contracts/bar.vy")
    bar = compile_from_file_input(
        bar_input, output_formats=output_formats, input_bundle=input_bundle
    )

    compile_code_results = {
        "contracts/bar.vy": bar,
        "contracts/library.vy": library,
        "contracts/foo.vy": foo,
    }

    output_json = compile_json(input_json)
    assert list(output_json["contracts"].keys()) == [
        "contracts/foo.vy",
        "contracts/library.vy",
        "contracts/bar.vy",
    ]

    assert sorted(output_json.keys()) == ["compiler", "contracts", "sources"]
    assert output_json["compiler"] == f"vyper-{vyper.__version__}"

    for source_id, contract_name in [(0, "foo"), (2, "library"), (3, "bar")]:
        path = f"contracts/{contract_name}.vy"
        data = compile_code_results[path]
        assert output_json["sources"][path] == {
            "id": source_id,
            "ast": data["ast_dict"]["ast"],
            "annotated_ast": data["annotated_ast_dict"]["ast"],
        }
        assert output_json["contracts"][path][contract_name] == {
            "abi": data["abi"],
            "devdoc": data["devdoc"],
            "interface": data["interface"],
            "ir": data["ir_dict"],
            "userdoc": data["userdoc"],
            "metadata": data["metadata"],
            "evm": {
                "bytecode": {
                    "object": data["bytecode"],
                    "opcodes": data["opcodes"],
                    "sourceMap": data["source_map"],
                },
                "deployedBytecode": {
                    "object": data["bytecode_runtime"],
                    "opcodes": data["opcodes_runtime"],
                    "sourceMap": data["source_map_runtime"],
                },
                "methodIdentifiers": data["method_identifiers"],
            },
        }


def test_compilation_targets(input_json):
    output_json = compile_json(input_json)
    assert list(output_json["contracts"].keys()) == [
        "contracts/foo.vy",
        "contracts/library.vy",
        "contracts/bar.vy",
    ]

    # omit library.vy
    input_json["settings"]["outputSelection"] = {"contracts/foo.vy": "*", "contracts/bar.vy": "*"}
    output_json = compile_json(input_json)

    assert list(output_json["contracts"].keys()) == ["contracts/foo.vy", "contracts/bar.vy"]


def test_different_outputs(input_bundle, input_json):
    input_json["settings"]["outputSelection"] = {
        "contracts/bar.vy": "*",
        "contracts/foo.vy": ["evm.methodIdentifiers"],
    }
    output_json = compile_json(input_json)
    assert list(output_json["contracts"].keys()) == ["contracts/bar.vy", "contracts/foo.vy"]

    assert sorted(output_json.keys()) == ["compiler", "contracts", "sources"]
    assert output_json["compiler"] == f"vyper-{vyper.__version__}"

    contracts = output_json["contracts"]

    foo = contracts["contracts/foo.vy"]["foo"]
    bar = contracts["contracts/bar.vy"]["bar"]
    assert sorted(bar.keys()) == ["abi", "devdoc", "evm", "interface", "ir", "metadata", "userdoc"]

    assert sorted(foo.keys()) == ["evm"]

    # check method_identifiers
    method_identifiers = compile_code(
        FOO_CODE,
        contract_path="contracts/foo.vy",
        output_formats=["method_identifiers"],
        input_bundle=input_bundle,
    )["method_identifiers"]
    assert foo["evm"]["methodIdentifiers"] == method_identifiers


def test_wrong_language():
    with pytest.raises(JSONError):
        compile_json({"language": "Solidity"})


def test_exc_handler_raises_syntax(input_json):
    input_json["sources"]["badcode.vy"] = {"content": BAD_SYNTAX_CODE}
    with pytest.raises(SyntaxException):
        compile_json(input_json)


def test_exc_handler_to_dict_syntax(input_json):
    input_json["sources"]["badcode.vy"] = {"content": BAD_SYNTAX_CODE}
    result = compile_json(input_json, exc_handler_to_dict)
    assert "errors" in result
    assert len(result["errors"]) == 1
    error = result["errors"][0]
    assert error["component"] == "compiler", error
    assert error["type"] == "SyntaxException"


def test_exc_handler_raises_compiler(input_json):
    input_json["sources"]["badcode.vy"] = {"content": BAD_COMPILER_CODE}
    with pytest.raises(TypeMismatch):
        compile_json(input_json)


def test_exc_handler_to_dict_compiler(input_json):
    input_json["sources"]["badcode.vy"] = {"content": BAD_COMPILER_CODE}
    result = compile_json(input_json, exc_handler_to_dict)
    assert sorted(result.keys()) == ["compiler", "errors"]
    assert result["compiler"] == f"vyper-{vyper.__version__}"
    assert len(result["errors"]) == 1
    error = result["errors"][0]
    assert error["component"] == "compiler"
    assert error["type"] == "TypeMismatch"


def test_source_ids_increment(input_json):
    input_json["settings"]["outputSelection"] = {"*": ["ast", "evm.deployedBytecode.sourceMap"]}
    result = compile_json(input_json)

    def get(filename, contractname):
        ast = result["sources"][filename]["ast"]
        ret = ast["source_id"]

        # grab it via source map to sanity check
        contract_info = result["contracts"][filename][contractname]["evm"]
        pc_ast_map = contract_info["deployedBytecode"]["sourceMap"]["pc_ast_map"]
        pc_item = next(iter(pc_ast_map.values()))
        source_id, node_id = pc_item
        assert ret == source_id

        return ret

    assert get("contracts/foo.vy", "foo") == 0
    assert get("contracts/library.vy", "library") == 2
    assert get("contracts/bar.vy", "bar") == 3


def test_relative_import_paths(input_json):
    input_json["sources"]["contracts/potato/baz/baz.vy"] = {"content": "from ... import foo"}
    input_json["sources"]["contracts/potato/baz/potato.vy"] = {"content": "from . import baz"}
    input_json["sources"]["contracts/potato/footato.vy"] = {"content": "from baz import baz"}
    compile_from_input_dict(input_json)



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import TypeMismatch

fail_list = [
    (
        """
@external
def foo() -> int128[2]:
    return [3,block.timestamp]
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo() -> int128[2]:
    return [block.timestamp - block.timestamp, block.timestamp]
    """,
        TypeMismatch,
    ),
    """
@external
def foo() -> decimal:
    x: int128 = as_wei_value(5, "finney")
    y: int128 = block.timestamp + 50
    return x // y
    """,
    (
        """
@external
def foo():
    x: Bytes[10] = slice(b"cow", -1, block.timestamp)
    """,
        TypeMismatch,
    ),
    """
@external
def foo():
    x: int128 = 7
    y: int128 = min(x, block.timestamp)
    """,
    """
a: HashMap[uint256, int128]

@external
def add_record():
    self.a[block.timestamp] = block.timestamp + 20
    """,
    """
a: HashMap[int128, int128]

@external
def add_record():
    self.a[block.timestamp] = block.timestamp + 20
    """,
    """
struct X:
    x: uint256
struct Y:
    y: int128
@external
def add_record():
    a: X = X(x=block.timestamp)
    b: Y = Y(y=5)
    a.x = b.y
    """,
    """
@external
def foo(inp: Bytes[10]) -> Bytes[3]:
    return slice(inp, block.timestamp, 6)
    """,
    (
        """
@external
def foo() -> int128:
    return block.fail
""",
        Exception,
    ),
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_block_fail(bad_code):
    if isinstance(bad_code, tuple):
        with raises(bad_code[1]):
            compiler.compile_code(bad_code[0])
    else:
        with raises(TypeMismatch):
            compiler.compile_code(bad_code)


valid_list = [
    """
a: HashMap[uint256, uint256]

@external
def add_record():
    self.a[block.timestamp] = block.timestamp + 20
    """,
    """
@external
def foo() -> uint256:
    x: uint256 = as_wei_value(5, "finney")
    y: uint256 = block.timestamp + 50 - block.timestamp
    return x // y
    """,
    """
@external
def foo() -> uint256[2]:
    return [block.timestamp + 86400, block.timestamp]
    """,
    """
@external
def foo():
    y: uint256 = min(block.timestamp + 30, block.timestamp + 50)
    """,
    """
struct X:
    x: uint256
@external
def add_record():
    a: X = X(x=block.timestamp)
    a.x = block.gaslimit
    a.x = block.basefee
    a.x = 5
    """,
    """
@external
def foo():
    x: uint256 = block.difficulty + 185
    if tx.origin == self:
        y: Bytes[35] = concat(block.prevhash, b"dog")
    """,
    """
@external
def foo():
    x: bytes32 = block.prevrandao
    if tx.origin == self:
        y: Bytes[35] = concat(block.prevhash, b"dog")
    """,
    """
@external
def foo() -> uint256:
    return tx.gasprice
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_block_success(good_code):
    assert compiler.compile_code(good_code) is not None


valid_list = [
    """
@external
def foo() -> uint256:
    return block.blobbasefee
    """,
    """
@external
def foo() -> uint256:
    a: uint256 = 5
    a = block.blobbasefee
    return a
    """,
    """
@external
def foo() -> uint256:
    a: uint256 = block.blobbasefee
    return a
    """,
]


@pytest.mark.requires_evm_version("cancun")
@pytest.mark.parametrize("good_code", valid_list)
def test_block_blob_success(good_code):
    assert compiler.compile_code(good_code) is not None



def test_concat(get_contract):
    test_concat = """
@external
def foo2(input1: Bytes[50], input2: Bytes[50]) -> Bytes[1000]:
    return concat(input1, input2)

@external
def foo3(input1: Bytes[50], input2: Bytes[50], input3: Bytes[50]) -> Bytes[1000]:
    return concat(input1, input2, input3)
    """

    c = get_contract(test_concat)
    assert c.foo2(b"h", b"orse") == b"horse"
    assert c.foo2(b"h", b"") == b"h"
    assert c.foo2(b"", b"") == b""
    assert c.foo2(b"", b"orse") == b"orse"
    assert c.foo3(b"Buffalo", b" ", b"buffalo") == b"Buffalo buffalo"
    assert c.foo2(b"\x36", b"\x35" * 32) == b"\x36" + b"\x35" * 32
    assert c.foo2(b"\x36" * 48, b"\x35" * 32) == b"\x36" * 48 + b"\x35" * 32
    assert (
        c.foo3(b"horses" * 4, b"mice" * 7, b"crows" * 10)
        == b"horses" * 4 + b"mice" * 7 + b"crows" * 10
    )  # noqa: E501
    print("Passed simple concat test")


def test_concat2(get_contract):
    test_concat2 = """
@external
def foo(inp: Bytes[50]) -> Bytes[1000]:
    x: Bytes[50] = inp
    return concat(x, inp, x, inp, x, inp, x, inp, x, inp)
    """

    c = get_contract(test_concat2)
    assert c.foo(b"horse" * 9 + b"vyper") == (b"horse" * 9 + b"vyper") * 10
    print("Passed second concat test")


def test_crazy_concat_code(get_contract):
    crazy_concat_code = """
y: Bytes[10]

@external
def krazykonkat(z: Bytes[10]) -> Bytes[25]:
    x: Bytes[3] = b"cow"
    self.y = b"horse"
    return concat(x, b" ", self.y, b" ", z)
    """

    c = get_contract(crazy_concat_code)

    assert c.krazykonkat(b"moose") == b"cow horse moose"

    print("Passed third concat test")


def test_concat_buffer(get_contract):
    # GHSA-2q8v-3gqq-4f8p
    code = """
@internal
def bar() -> uint256:
    sss: String[2] = concat("a", "b")
    return 1


@external
def foo() -> int256:
    a: int256 = -1
    b: uint256 = self.bar()
    return a
    """
    c = get_contract(code)
    assert c.foo() == -1


def test_concat_buffer2(get_contract):
    # GHSA-2q8v-3gqq-4f8p
    code = """
i: immutable(int256)

@deploy
def __init__():
    i = -1
    s: String[2] = concat("a", "b")

@external
def foo() -> int256:
    return i
    """
    c = get_contract(code)
    assert c.foo() == -1


def test_concat_buffer3(get_contract):
    # GHSA-2q8v-3gqq-4f8p
    code = """
s: String[1]
s2: String[33]
s3: String[34]

@deploy
def __init__():
    self.s = "a"
    self.s2 = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" # 33*'a'

@internal
def bar() -> uint256:
    self.s3 = concat(self.s, self.s2)
    return 1

@external
def foo() -> int256:
    i: int256 = -1
    b: uint256 = self.bar()
    return i
    """
    c = get_contract(code)
    assert c.foo() == -1


def test_concat_bytes32(get_contract):
    test_concat_bytes32 = """
@external
def sandwich(inp: Bytes[100], inp2: bytes32) -> Bytes[164]:
    return concat(inp2, inp, inp2)

@external
def fivetimes(inp: bytes32) -> Bytes[160]:
    return concat(inp, inp, inp, inp, inp)
    """

    c = get_contract(test_concat_bytes32)
    assert c.sandwich(b"cow", b"\x35" * 32) == b"\x35" * 32 + b"cow" + b"\x35" * 32, c.sandwich(
        b"cow", b"\x35" * 32
    )  # noqa: E501
    assert c.sandwich(b"", b"\x46" * 32) == b"\x46" * 64
    assert c.sandwich(b"\x57" * 95, b"\x57" * 32) == b"\x57" * 159
    assert c.sandwich(b"\x57" * 96, b"\x57" * 32) == b"\x57" * 160
    assert c.sandwich(b"\x57" * 97, b"\x57" * 32) == b"\x57" * 161
    assert c.fivetimes(b"mongoose" * 4) == b"mongoose" * 20

    print("Passed concat bytes32 test")


def test_konkat_code(get_contract):
    konkat_code = """
ecks: bytes32

@external
def foo(x: bytes32, y: bytes32) -> Bytes[64]:
    self.ecks = x
    return concat(self.ecks, y)

@external
def goo(x: bytes32, y: bytes32) -> Bytes[64]:
    self.ecks = x
    return concat(self.ecks, y)

@external
def hoo(x: bytes32, y: bytes32) -> Bytes[64]:
    return concat(x, y)
    """

    c = get_contract(konkat_code)
    assert c.foo(b"\x35" * 32, b"\x00" * 32) == b"\x35" * 32 + b"\x00" * 32
    assert c.goo(b"\x35" * 32, b"\x00" * 32) == b"\x35" * 32 + b"\x00" * 32
    assert c.hoo(b"\x35" * 32, b"\x00" * 32) == b"\x35" * 32 + b"\x00" * 32

    print("Passed second concat tests")


def test_small_output(get_contract):
    code = """
@external
def small_output(a: String[5], b: String[4]) -> String[9]:
    c: String[9] = concat(a, b)
    return c
    """
    c = get_contract(code)
    assert c.small_output("abcde", "fghi") == "abcdefghi"
    assert c.small_output("", "") == ""


def test_small_bytes(get_contract):
    # TODO maybe use parametrization or hypothesis for the examples
    code = """
@external
def small_bytes1(a: bytes1, b: Bytes[2]) -> Bytes[3]:
    return concat(a, b)

@external
def small_bytes2(a: Bytes[1], b: bytes2) -> Bytes[3]:
    return concat(a, b)

@external
def small_bytes3(a: bytes4, b: bytes32) -> Bytes[36]:
    return concat(a, b)

@external
def small_bytes4(a: bytes8, b: Bytes[32], c: bytes8) -> Bytes[48]:
    return concat(a, b, c)
    """
    contract = get_contract(code)

    i = 0

    def bytes_for_len(n):
        nonlocal i
        # bytes constructor with state
        # (so we don't keep generating the same string)
        xs = []
        for _ in range(n):
            i += 1
            i %= 256
            xs.append(i)
        return bytes(xs)

    a, b = bytes_for_len(1), bytes_for_len(2)
    assert contract.small_bytes1(a, b) == a + b

    a, b = bytes_for_len(1), bytes_for_len(1)
    assert contract.small_bytes1(a, b) == a + b

    a, b = bytes_for_len(1), bytes_for_len(2)
    assert contract.small_bytes2(a, b) == a + b

    a, b = b"", bytes_for_len(2)
    assert contract.small_bytes2(a, b) == a + b

    a, b = bytes_for_len(4), bytes_for_len(32)
    assert contract.small_bytes3(a, b) == a + b

    a, b, c = bytes_for_len(8), bytes_for_len(32), bytes_for_len(8)
    assert contract.small_bytes4(a, b, c) == a + b + c

    a, b, c = bytes_for_len(8), bytes_for_len(1), bytes_for_len(8)
    assert contract.small_bytes4(a, b, c) == a + b + c

    a, b, c = bytes_for_len(8), bytes_for_len(0), bytes_for_len(8)
    assert contract.small_bytes4(a, b, c) == a + b + c



from .. import types  # break a dependency cycle.
from .global_ import validate_compilation_target
from .module import analyze_module

__all__ = [validate_compilation_target, analyze_module]  # type: ignore[misc]



import re
from dataclasses import dataclass
from typing import Optional, Tuple

from asttokens import LineNumbers

from vyper.ast import nodes as vy_ast
from vyper.exceptions import NatSpecSyntaxException

SINGLE_FIELDS = ("title", "author", "license", "notice", "dev")
PARAM_FIELDS = ("param", "return")
USERDOCS_FIELDS = ("notice",)


@dataclass
class NatspecOutput:
    userdoc: dict
    devdoc: dict


def parse_natspec(annotated_vyper_module: vy_ast.Module) -> NatspecOutput:
    """
    Parses NatSpec documentation from a contract.

    Arguments
    ---------
    annotated_vyper_module: Module
        Module-level vyper ast node.
    interface_codes: Dict, optional
        Dict containing relevant data for any import statements related to
        this contract.

    Returns
    -------
    dict
        NatSpec user documentation
    dict
        NatSpec developer documentation
    """
    from vyper.semantics.types.function import FunctionVisibility

    userdoc, devdoc = {}, {}
    source: str = annotated_vyper_module.full_source_code

    docstring = annotated_vyper_module.get("doc_string.value")
    if docstring:
        devdoc.update(_parse_docstring(source, docstring, ("param", "return")))
        if "notice" in devdoc:
            userdoc["notice"] = devdoc.pop("notice")

    for node in [i for i in annotated_vyper_module.body if i.get("doc_string.value")]:
        docstring = node.doc_string.value
        func_type = node._metadata["func_type"]
        if func_type.visibility != FunctionVisibility.EXTERNAL:
            continue

        if isinstance(node.returns, vy_ast.Tuple):
            ret_len = len(node.returns.elements)
        elif node.returns:
            ret_len = 1
        else:
            ret_len = 0

        args = tuple(i.arg for i in node.args.args)
        invalid_fields = ("title", "license")
        fn_natspec = _parse_docstring(source, docstring, invalid_fields, args, ret_len)
        for method_id in func_type.method_ids:
            if "notice" in fn_natspec:
                userdoc.setdefault("methods", {})[method_id] = {"notice": fn_natspec.pop("notice")}
            if fn_natspec:
                devdoc.setdefault("methods", {})[method_id] = fn_natspec

    return NatspecOutput(userdoc=userdoc, devdoc=devdoc)


def _parse_docstring(
    source: str,
    docstring: str,
    invalid_fields: Tuple,
    params: Optional[Tuple] = None,
    return_length: int = 0,
) -> dict:
    natspec: dict = {}
    if params is None:
        params = tuple()

    line_no = LineNumbers(source)
    start = source.index(docstring)

    translate_map = {"return": "returns", "dev": "details", "param": "params"}

    pattern = r"(?:^|\n)\s*@(\S+)\s*([\s\S]*?)(?=\n\s*@\S|\s*$)"

    for match in re.finditer(pattern, docstring):
        tag, value = match.groups()
        err_args = (source, *line_no.offset_to_line(start + match.start(1)))

        if tag not in SINGLE_FIELDS + PARAM_FIELDS and not tag.startswith("custom:"):
            raise NatSpecSyntaxException(f"Unknown NatSpec field '@{tag}'", *err_args)
        if tag in invalid_fields:
            raise NatSpecSyntaxException(
                f"'@{tag}' is not a valid field for this docstring", *err_args
            )

        if not value or value.startswith("@"):
            raise NatSpecSyntaxException(f"No description given for tag '@{tag}'", *err_args)

        if tag not in PARAM_FIELDS:
            if tag in natspec:
                raise NatSpecSyntaxException(f"Duplicate NatSpec field '@{tag}'", *err_args)
            natspec[translate_map.get(tag, tag)] = " ".join(value.split())
            continue

        tag = translate_map.get(tag, tag)
        natspec.setdefault(tag, {})

        if tag == "params":
            try:
                key, value = value.split(maxsplit=1)
            except ValueError as exc:
                raise NatSpecSyntaxException(
                    f"No description given for parameter '{value}'", *err_args
                ) from exc
            if key not in params:
                raise NatSpecSyntaxException(f"Method has no parameter '{key}'", *err_args)

        elif tag == "returns":
            if not return_length:
                raise NatSpecSyntaxException("Method does not return any values", *err_args)
            if len(natspec["returns"]) >= return_length:
                raise NatSpecSyntaxException(
                    "Number of documented return values exceeds actual number", *err_args
                )
            key = f"_{len(natspec['returns'])}"

        if key in natspec[tag]:
            raise NatSpecSyntaxException(f"Parameter '{key}' documented more than once", *err_args)
        natspec[tag][key] = " ".join(value.split())

    if not natspec:
        natspec["notice"] = " ".join(docstring.split())
    elif not docstring.strip().startswith("@"):
        raise NatSpecSyntaxException(
            "NatSpec docstring opens with untagged comment", source, *line_no.offset_to_line(start)
        )

    return natspec



import hypothesis.strategies as st
import pytest
from hypothesis import given, settings

import vyper.utils as utils
from vyper.codegen.jumptable_utils import (
    generate_dense_jumptable_info,
    generate_sparse_jumptable_buckets,
)
from vyper.compiler.settings import OptimizationLevel


def test_dense_selector_table_empty_buckets(get_contract):
    # some special combination of selectors which can result in
    # some empty bucket being returned from _mk_buckets (that is,
    # len(_mk_buckets(..., n_buckets)) != n_buckets
    code = """
@external
def aX61QLPWF()->uint256:
    return 1
@external
def aQHG0P2L1()->uint256:
    return 2
@external
def a2G8ME94W()->uint256:
    return 3
@external
def a0GNA21AY()->uint256:
    return 4
@external
def a4U1XA4T5()->uint256:
    return 5
@external
def aAYLMGOBZ()->uint256:
    return 6
@external
def a0KXRLHKE()->uint256:
    return 7
@external
def aDQS32HTR()->uint256:
    return 8
@external
def aP4K6SA3S()->uint256:
    return 9
@external
def aEB94ZP5S()->uint256:
    return 10
@external
def aTOIMN0IM()->uint256:
    return 11
@external
def aXV2N81OW()->uint256:
    return 12
@external
def a66PP6Y5X()->uint256:
    return 13
@external
def a5MWMTEWN()->uint256:
    return 14
@external
def a5ZFST4Z8()->uint256:
    return 15
@external
def aR13VXULX()->uint256:
    return 16
@external
def aWITH917Y()->uint256:
    return 17
@external
def a59NP6C5O()->uint256:
    return 18
@external
def aJ02590EX()->uint256:
    return 19
@external
def aUAXAAUQ8()->uint256:
    return 20
@external
def aWR1XNC6J()->uint256:
    return 21
@external
def aJABKZOKH()->uint256:
    return 22
@external
def aO1TT0RJT()->uint256:
    return 23
@external
def a41442IOK()->uint256:
    return 24
@external
def aMVXV9FHQ()->uint256:
    return 25
@external
def aNN0KJDZM()->uint256:
    return 26
@external
def aOX965047()->uint256:
    return 27
@external
def a575NX2J3()->uint256:
    return 28
@external
def a16EN8O7W()->uint256:
    return 29
@external
def aSZXLFF7O()->uint256:
    return 30
@external
def aQKQCIPH9()->uint256:
    return 31
@external
def aIP8021DL()->uint256:
    return 32
@external
def aQAV0HSHX()->uint256:
    return 33
@external
def aZVPAD745()->uint256:
    return 34
@external
def aJYBSNST4()->uint256:
    return 35
@external
def aQGWC4NYQ()->uint256:
    return 36
@external
def aFMBB9CXJ()->uint256:
    return 37
@external
def aYWM7ZUH1()->uint256:
    return 38
@external
def aJAZONIX1()->uint256:
    return 39
@external
def aQZ1HJK0H()->uint256:
    return 40
@external
def aKIH9LOUB()->uint256:
    return 41
@external
def aF4ZT80XL()->uint256:
    return 42
@external
def aYQD8UKR5()->uint256:
    return 43
@external
def aP6NCCAI4()->uint256:
    return 44
@external
def aY92U2EAZ()->uint256:
    return 45
@external
def aHMQ49D7P()->uint256:
    return 46
@external
def aMC6YX8VF()->uint256:
    return 47
@external
def a734X6YSI()->uint256:
    return 48
@external
def aRXXPNSMU()->uint256:
    return 49
@external
def aL5XKDTGT()->uint256:
    return 50
@external
def a86V1Y18A()->uint256:
    return 51
@external
def aAUM8PL5J()->uint256:
    return 52
@external
def aBAEC1ERZ()->uint256:
    return 53
@external
def a1U1VA3UE()->uint256:
    return 54
@external
def aC9FGVAHC()->uint256:
    return 55
@external
def aWN81WYJ3()->uint256:
    return 56
@external
def a3KK1Y07J()->uint256:
    return 57
@external
def aAZ6P6OSG()->uint256:
    return 58
@external
def aWP5HCIB3()->uint256:
    return 59
@external
def aVEK161C5()->uint256:
    return 60
@external
def aY0Q3O519()->uint256:
    return 61
@external
def aDHHHFIAE()->uint256:
    return 62
@external
def aGSJBCZKQ()->uint256:
    return 63
@external
def aZQQIUDHY()->uint256:
    return 64
@external
def a12O9QDH5()->uint256:
    return 65
@external
def aRQ1178XR()->uint256:
    return 66
@external
def aDT25C832()->uint256:
    return 67
@external
def aCSB01C4E()->uint256:
    return 68
@external
def aYGBPKZSD()->uint256:
    return 69
@external
def aP24N3EJ8()->uint256:
    return 70
@external
def a531Y9X3C()->uint256:
    return 71
@external
def a4727IKVS()->uint256:
    return 72
@external
def a2EX1L2BS()->uint256:
    return 73
@external
def a6145RN68()->uint256:
    return 74
@external
def aDO1ZNX97()->uint256:
    return 75
@external
def a3R28EU6M()->uint256:
    return 76
@external
def a9BFC867L()->uint256:
    return 77
@external
def aPL1MBGYC()->uint256:
    return 78
@external
def aI6H11O48()->uint256:
    return 79
@external
def aX0248DZY()->uint256:
    return 80
@external
def aE4JBUJN4()->uint256:
    return 81
@external
def aXBDB2ZBO()->uint256:
    return 82
@external
def a7O7MYYHL()->uint256:
    return 83
@external
def aERFF4PB6()->uint256:
    return 84
@external
def aJCUBG6TJ()->uint256:
    return 85
@external
def aQ5ELXM0F()->uint256:
    return 86
@external
def aWDT9UQVV()->uint256:
    return 87
@external
def a7UU40DJK()->uint256:
    return 88
@external
def aH01IT5VS()->uint256:
    return 89
@external
def aSKYTZ0FC()->uint256:
    return 90
@external
def aNX5LYRAW()->uint256:
    return 91
@external
def aUDKAOSGG()->uint256:
    return 92
@external
def aZ86YGAAO()->uint256:
    return 93
@external
def aIHWQGKLO()->uint256:
    return 94
@external
def aKIKFLAR9()->uint256:
    return 95
@external
def aCTPE0KRS()->uint256:
    return 96
@external
def aAD75X00P()->uint256:
    return 97
@external
def aDROUEF2F()->uint256:
    return 98
@external
def a8CDIF6YN()->uint256:
    return 99
@external
def aD2X7TM83()->uint256:
    return 100
@external
def a3W5UUB4L()->uint256:
    return 101
@external
def aG4MOBN4B()->uint256:
    return 102
@external
def aPRS0MSG7()->uint256:
    return 103
@external
def aKN3GHBUR()->uint256:
    return 104
@external
def aGE435RHQ()->uint256:
    return 105
@external
def a4E86BNFE()->uint256:
    return 106
@external
def aYDG928YW()->uint256:
    return 107
@external
def a2HFP5GQE()->uint256:
    return 108
@external
def a5DPMVXKA()->uint256:
    return 109
@external
def a3OFVC3DR()->uint256:
    return 110
@external
def aK8F62DAN()->uint256:
    return 111
@external
def aJS9EY3U6()->uint256:
    return 112
@external
def aWW789JQH()->uint256:
    return 113
@external
def a8AJJN3YR()->uint256:
    return 114
@external
def a4D0MUIDU()->uint256:
    return 115
@external
def a35W41JQR()->uint256:
    return 116
@external
def a07DQOI1E()->uint256:
    return 117
@external
def aFT43YNCT()->uint256:
    return 118
@external
def a0E75I8X3()->uint256:
    return 119
@external
def aT6NXIRO4()->uint256:
    return 120
@external
def aXB2UBAKQ()->uint256:
    return 121
@external
def aHWH55NW6()->uint256:
    return 122
@external
def a7TCFE6C2()->uint256:
    return 123
@external
def a8XYAM81I()->uint256:
    return 124
@external
def aHQTQ4YBY()->uint256:
    return 125
@external
def aGCZEHG6Y()->uint256:
    return 126
@external
def a6LJTKIW0()->uint256:
    return 127
@external
def aBDIXTD9S()->uint256:
    return 128
@external
def aCB83G21P()->uint256:
    return 129
@external
def aZC525N4K()->uint256:
    return 130
@external
def a40LC94U6()->uint256:
    return 131
@external
def a8X9TI93D()->uint256:
    return 132
@external
def aGUG9CD8Y()->uint256:
    return 133
@external
def a0LAERVAY()->uint256:
    return 134
@external
def aXQ0UEX19()->uint256:
    return 135
@external
def aKK9C7NE7()->uint256:
    return 136
@external
def aS2APW8UE()->uint256:
    return 137
@external
def a65NT07MM()->uint256:
    return 138
@external
def aGRMT6ZW5()->uint256:
    return 139
@external
def aILR4U1Z()->uint256:
    return 140
    """
    c = get_contract(code)

    assert c.aX61QLPWF() == 1  # will revert if the header section is misaligned


@given(
    n_methods=st.integers(min_value=1, max_value=100),
    seed=st.integers(min_value=0, max_value=2**64 - 1),
)
@pytest.mark.fuzzing
@settings(max_examples=10)
def test_sparse_jumptable_probe_depth(n_methods, seed):
    sigs = [f"foo{i + seed}()" for i in range(n_methods)]
    _, buckets = generate_sparse_jumptable_buckets(sigs)
    bucket_sizes = [len(bucket) for bucket in buckets.values()]

    # generally bucket sizes should be bounded at around 4, but
    # just test that they don't get really out of hand
    assert max(bucket_sizes) <= 8

    # generally mean bucket size should be around 1.6, here just
    # test they don't get really out of hand
    assert sum(bucket_sizes) / len(bucket_sizes) <= 4


@given(
    n_methods=st.integers(min_value=4, max_value=100),
    seed=st.integers(min_value=0, max_value=2**64 - 1),
)
@pytest.mark.fuzzing
@settings(max_examples=10)
def test_dense_jumptable_bucket_size(n_methods, seed):
    sigs = [f"foo{i + seed}()" for i in range(n_methods)]
    n = len(sigs)
    buckets = generate_dense_jumptable_info(sigs)
    n_buckets = len(buckets)

    # generally should be around 14 buckets per 100 methods, here
    # we test they don't get really out of hand
    assert n_buckets / n < 0.4 or n < 10


@st.composite
def generate_methods(draw, max_calldata_bytes):
    max_default_args = draw(st.integers(min_value=0, max_value=4))
    default_fn_mutability = draw(st.sampled_from(["", "@pure", "@view", "@nonpayable", "@payable"]))

    return (
        max_default_args,
        default_fn_mutability,
        draw(
            st.lists(
                st.tuples(
                    # function id:
                    st.integers(min_value=0),
                    # mutability:
                    st.sampled_from(["@pure", "@view", "@nonpayable", "@payable"]),
                    # n calldata words:
                    st.integers(min_value=0, max_value=max_calldata_bytes // 32),
                    # n bytes to strip from calldata
                    st.integers(min_value=1, max_value=4),
                    # n default args
                    st.integers(min_value=0, max_value=max_default_args),
                ),
                unique_by=lambda x: x[0],
                min_size=1,
                max_size=100,
            )
        ),
    )


@pytest.mark.parametrize("opt_level", list(OptimizationLevel))
# dense selector table packing boundaries at 256 and 65336
@pytest.mark.parametrize("max_calldata_bytes", [255, 256, 65336])
@pytest.mark.fuzzing
def test_selector_table_fuzz(max_calldata_bytes, opt_level, env, get_contract, tx_failed, get_logs):
    def abi_sig(func_id, calldata_words, n_default_args):
        params = [] if not calldata_words else [f"uint256[{calldata_words}]"]
        params.extend(["uint256"] * n_default_args)
        paramstr = ",".join(params)
        return f"foo{func_id}({paramstr})"

    def generate_func_def(func_id, mutability, calldata_words, n_default_args):
        arglist = [] if not calldata_words else [f"x: uint256[{calldata_words}]"]
        for j in range(n_default_args):
            arglist.append(f"x{j}: uint256 = 0")
        args = ", ".join(arglist)
        _log_return = f"log _Return({func_id})" if mutability == "@payable" else ""

        return f"""
@external
{mutability}
def foo{func_id}({args}) -> uint256:
    {_log_return}
    return {func_id}
    """

    @given(_input=generate_methods(max_calldata_bytes))
    @settings(max_examples=125)
    def _test(_input):
        max_default_args, default_fn_mutability, methods = _input

        func_defs = "\n".join(
            generate_func_def(func_id, mutability, calldata_words, n_default_args)
            for (func_id, mutability, calldata_words, _, n_default_args) in (methods)
        )

        if default_fn_mutability == "":
            default_fn_code = ""
        elif default_fn_mutability in ("@nonpayable", "@payable"):
            default_fn_code = f"""
@external
{default_fn_mutability}
def __default__():
    log CalledDefault()
            """
        else:
            # can't log from pure/view functions, just test that it returns
            default_fn_code = """
@external
def __default__():
    pass
            """

        code = f"""
event CalledDefault:
    pass

event _Return:
    val: uint256

{func_defs}

{default_fn_code}
        """

        c = get_contract(code, override_opt_level=opt_level)
        env.set_balance(env.deployer, 10**18)

        for func_id, mutability, n_calldata_words, n_strip_bytes, n_default_args in methods:
            funcname = f"foo{func_id}"
            func = getattr(c, funcname)

            for j in range(n_default_args + 1):
                args = [[1] * n_calldata_words] if n_calldata_words else []
                args.extend([1] * j)

                # check the function returns as expected
                assert func(*args) == func_id

                method_id = utils.method_id(abi_sig(func_id, n_calldata_words, j))

                argsdata = b"\x00" * (n_calldata_words * 32 + j * 32)

                # do payable check
                if mutability == "@payable":
                    func(*args, value=1)
                    (event,) = get_logs(c, "_Return")
                    assert event.args.val == func_id
                else:
                    hexstr = (method_id + argsdata).hex()
                    with tx_failed():
                        env.message_call(c.address, data=hexstr, value=1)

                # now do calldatasize check
                # strip some bytes
                calldata = (method_id + argsdata)[:-n_strip_bytes]
                hexstr = calldata.hex()
                tx_params = {"to": c.address, "data": hexstr}
                if n_calldata_words == 0 and j == 0:
                    # no args, hit default function
                    if default_fn_mutability == "":
                        with tx_failed():
                            env.message_call(**tx_params)
                    elif default_fn_mutability == "@payable":
                        # we should be able to send eth to it
                        tx_params["value"] = 1
                        env.message_call(**tx_params)
                        logs = get_logs(c, "CalledDefault")
                        assert len(logs) == 1
                    else:
                        env.message_call(**tx_params)

                        # note: can't emit logs from view/pure functions,
                        # so the logging is not tested.
                        if default_fn_mutability == "@nonpayable":
                            logs = get_logs(c, "CalledDefault")
                            assert len(logs) == 1

                        # check default function reverts
                        tx_params["value"] = 1
                        with tx_failed():
                            env.message_call(**tx_params)
                else:
                    with tx_failed():
                        env.message_call(**tx_params)

    _test()



import pytest

from vyper import compiler
from vyper.exceptions import InvalidType, StructureException

fail_list = [
    """
x[5] = 4
    """,
    """
send(0x1234567890123456789012345678901234567890, 5)
    """,
    """
send(0x1234567890123456789012345678901234567890, 5)
    """,
    """
x: [bar, baz]
    """,
    """
x: [bar(int128), baz(baffle)]
    """,
    """
x: int128
@external
@view(123)
def foo() -> int128:
    pass
    """,
    """
@external
def foo():
    throe
    """,
    """
@external
def foo() -> int128:
    x: address = 0x1234567890123456789012345678901234567890
    return x.balance()
    """,
    """
@external
def foo() -> int128:
    x: address = 0x1234567890123456789012345678901234567890
    return x.codesize()
    """,
    """
struct X:
    int128[5]: int128[7]
    """,
    """
@external
def foo():
    true: int128 = 3
    """,
    """
n: HashMap[uint256, bool][3]
    """,
    """
a: constant(uint256) = 3
n: public(HashMap[uint256, uint256][a])
    """,
    """
a: immutable(uint256)
n: public(HashMap[uint256, bool][a])

@deploy
def __init__():
    a = 3
    """,
    """
n: HashMap[uint256, bool][3][3]
    """,
    """
m1: HashMap[uint8, uint8]
m2: HashMap[uint8, uint8]

@deploy
def __init__():
    self.m1 = self.m2
    """,
    """
m1: HashMap[uint8, uint8]

@deploy
def __init__():
    self.m1 = 234
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_invalid_type_exception(bad_code):
    with pytest.raises((StructureException, InvalidType)):
        compiler.compile_code(bad_code)


del_fail_list = [
    """
x: int128(address)
    """,
    """
x: int128(2 ** 2)
    """,
    """
# invalid interface declaration (pass)
interface Bar:
    def set_lucky(arg1: int128): pass
    """,
    """
interface Bar:
# invalid interface declaration (assignment)
    def set_lucky(arg1: int128):
        arg1 = 1
        arg1 = 3
    """,
]



import pytest

from vyper.exceptions import StructureException


def test_continue1(get_contract):
    code = """
@external
def foo() -> bool:
    for i: uint256 in range(2):
        continue
        return False
    return True
"""
    c = get_contract(code)
    assert c.foo()


def test_continue2(get_contract):
    code = """
@external
def foo() -> int128:
    x: int128 = 0
    for i: int128 in range(3):
        x += 1
        continue
        x -= 1
    return x
"""
    c = get_contract(code)
    assert c.foo() == 3


def test_continue3(get_contract):
    code = """
@external
def foo() -> int128:
    x: int128 = 0
    for i: int128 in range(3):
        x += i
        continue
    return x
"""
    c = get_contract(code)
    assert c.foo() == 3


def test_continue4(get_contract):
    code = """
@external
def foo() -> int128:
    x: int128 = 0
    for i: int128 in range(6):
        if i % 2 == 0:
            continue
        x += 1
    return x
"""
    c = get_contract(code)
    assert c.foo() == 3


fail_list = [
    (
        """
@external
def foo():
    a: uint256 = 3
    continue
    """,
        StructureException,
    ),
    (
        """
@external
def foo():
    if True:
        continue
    """,
        StructureException,
    ),
    (
        """
@external
def foo():
    for i: uint256 in [1, 2, 3]:
        b: uint256 = i
    if True:
        continue
    """,
        StructureException,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_block_fail(assert_compile_failed, get_contract, bad_code, exc):
    assert_compile_failed(lambda: get_contract(bad_code), exc)



def test_comment_test(get_contract):
    comment_test = """
@external
def foo() -> int128:
    # Returns 3
    return 3
    """

    c = get_contract(comment_test)
    assert c.foo() == 3
    print("Passed comment test")



from vyper.ast.parse import parse_to_ast


def test_ast_equal():
    code = """
@external
def test() -> int128:
    a: uint256 = 100
    return 123
    """

    ast1 = parse_to_ast(code)
    ast2 = parse_to_ast("\n   \n" + code + "\n\n")

    assert ast1 == ast2


def test_ast_unequal():
    code1 = """
@external
def test() -> int128:
    a: uint256 = 100
    return 123
    """
    code2 = """
@external
def test() -> int128:
    a: uint256 = 100
    return 121
    """

    ast1 = parse_to_ast(code1)
    ast2 = parse_to_ast(code2)

    assert ast1 != ast2
